import numpy as np
from scipy.linalg import hadamard
import scipy.special as sp
import scipy.stats as stats
import scipy.signal.windows as scipy_windows
import math
import numba



class MIMO_Functions:
  
  def sv(self,az,fd,W):
    NPulse=W.shape[0]
    M=W.shape[1]
    a=np.exp(1j*np.pi*np.arange(M)*np.sin(np.deg2rad(az)))
    d=np.exp(1j*2*np.pi*np.arange(NPulse)*fd)
    s = (W @ a)*d
    return s

  def plot_Angle_Doppler(self,W,st,ax):
    x = self.sv(20,0,W)
    azv = np.linspace(-90,90,100)
    fdv = np.linspace(-.5,.5,500)
    Im = np.zeros((azv.shape[0],fdv.shape[0]),dtype=np.float64)
    for i in range(azv.shape[0]):
      for j in range(fdv.shape[0]):
        Im[i,j] = (np.abs(np.conj(self.sv(azv[i],fdv[j],W)).T@x))
    im = ax.imshow(Im.T, extent=(-90, 90, -0.5, 0.5), aspect='auto', cmap='viridis')  # Added cmap for better color representation
    ax.set_title(st)
    cbar = plt.colorbar(im, ax=ax)
    # cbar.set_label('Log-Power (dB)')

  def AD_matrix(self,NPulse,M,tech='TDM'):
    W = np.zeros((NPulse,M),dtype=np.complex128)
    match tech:
      case "DDM":
        for p in range(NPulse):
          for m in range(M):
            W[p,m]=np.exp(1j*2*np.pi*m*p/M)
      case "DDM2":
        for p in range(NPulse):
          for m in range(M):
            W[p,m]=np.exp(1j*2*np.pi*m*p/NPulse)
      case "DDM_RandPhase":
        for p in range(NPulse):
          for m in range(M):
            W[p,m]=np.exp(1j*2*np.pi*np.random.rand())
      case "TDM":
        for p in range(NPulse):
          for m in range(M):
            W[p,m]=0
            if p%M == m:
              W[p,m]=1
      case "BPM":
        H = hadamard(M, dtype=complex)
        for p in range(NPulse):
          for m in range(M):
            W[p,m]=H[p%M,m]
      case _:
          print("Not Valid!")
    return W
  

def steeringVector_position(specifications,p0,Coherent_Distributed=False):
  global_location_TX,global_location_RX,global_location_Center = specifications['global_location_TX_RX_Center']
  sv = np.zeros((len(global_location_TX)*len(global_location_RX),1),dtype=complex)
  k=0
  for itx , txPos in enumerate(global_location_TX):
    dtx = np.linalg.norm(p0-txPos)
    for irx , rxPos in enumerate(global_location_RX):
      drx = np.linalg.norm(p0-rxPos)
      sv[k,0]=np.exp(1j*2*np.pi/specifications['Lambda']*(dtx+drx))
      k+=1
  if Coherent_Distributed == False:
    sv/=sv[0]
  return sv

def update_block_diag_matrix(Z, vec):
    if Z is None:
        return vec
    rowsZ, colsZ = Z.shape
    newZ = np.zeros((rowsZ + vec.shape[0], colsZ + vec.shape[1]),dtype=complex)
    newZ[:rowsZ, :colsZ] = Z
    newZ[rowsZ:, colsZ:] = vec
    return newZ

def Distributed_MIMO_steeringMatrix_position(RadarSpecifications,p0,allSuites=1):
  if allSuites==1:
    steeringMatrix = None
    for radarSpecifications in RadarSpecifications:
      for specifications in radarSpecifications:
        steeringMatrix = update_block_diag_matrix(steeringMatrix,steeringVector_position(specifications,p0))
    return steeringMatrix
  elif allSuites==2:
    steeringMatrix=[]
    for radarSpecifications in RadarSpecifications:
      steeringMatrix0 = None
      for specifications in radarSpecifications:
        steeringMatrix0 = update_block_diag_matrix(steeringMatrix0,steeringVector_position(specifications,p0))
      steeringMatrix.append(steeringMatrix0)
    return steeringMatrix
  elif allSuites==3:
    steeringMatrix=[]
    for radarSpecifications in RadarSpecifications:
      for specifications in radarSpecifications:
        steeringMatrix.append(steeringVector_position(specifications,p0))
    return steeringMatrix
  
def generate_covariance_matrix(N, rho):
    covariance_matrix = np.zeros((N, N), dtype=complex)
    for i in range(N):
        for j in range(N):
            # covariance_matrix[i, j] = np.exp(-spread * abs(i - j))
            covariance_matrix[i, j] = (rho**abs(i - j))*np.exp(1j* np.pi/2 *(i - j))
    return covariance_matrix
  
def generate_covariance_matrix_Distributed(Nv, rhov,c=0):
  covariance_matrix = None
  for i in range(len(Nv)):
    cm = generate_covariance_matrix(Nv[i], rhov[i])
    covariance_matrix = update_block_diag_matrix(covariance_matrix,cm)
  for i in range(covariance_matrix.shape[0]):
    for j in range(i):
      if covariance_matrix[i,j]==0:
        a= c * np.exp(1j*np.random.rand()*2*np.pi)
        covariance_matrix[i,j]=a
        covariance_matrix[j,i]=np.conj(a)
  return covariance_matrix

#@numba.jit(nopython=True)
def generate_complex_gaussian_samples(cholesky_covariance_matrix, K):
    N = cholesky_covariance_matrix.shape[0]
    real_part = np.random.randn(N, K)
    imag_part = np.random.randn(N, K)
    Z = (real_part + 1j * imag_part) / np.sqrt(2)
    samples = cholesky_covariance_matrix @ Z
    return samples

def pfaTHR_tH0(tH0,pfa_v):
  N_MC = tH0.shape[0]
  detector_THR = tH0[tH0.argsort()]
  indices = []
  Pfa_THR  = []
  for pfai in pfa_v:
      ind = round((1 - pfai) * (N_MC - 1))
      if ind not in indices:
          indices.append(ind)
          THR = detector_THR[ind]
          pfa = np.sum(tH0>THR)/N_MC
          Pfa_THR.append([pfa,THR])
  return np.array(Pfa_THR)

def ROC_tH1H0(tH0H1,pfa_v):
    N_MC = tH0H1.shape[0]
    detector_THR = tH0H1[:, 0]   
    detector_THR = detector_THR[detector_THR.argsort()]
    indices = []
    Pd_Pfa_THR  = []
    for pfai in pfa_v:
        ind = round((1 - pfai) * (N_MC - 1))
        if ind not in indices:
            indices.append(ind)
            THR = detector_THR[ind]
            pfa = np.sum(tH0H1[:, 0]>THR)/N_MC
            pd  = np.sum(tH0H1[:, 1]>THR)/N_MC
            Pd_Pfa_THR.append([pd,pfa,THR])
    return np.array(Pd_Pfa_THR)
  
def ROC_tH1(tH1,pfa_v,THR_v):
    Pd_Pfa  = []
    for i in range(pfa_v.shape[0]):
        pd  = np.sum(tH1>THR_v[i])/tH1.shape[0]
        Pd_Pfa.append([pd,pfa_v[i]])
    return np.array(Pd_Pfa)
#@numba.jit(nopython=True)
def array_detector_ABC(A,B,C):
  return A@B@C

def array_detector_ABC_inv(A,B,C):
  return np.linalg.pinv(array_detector_ABC(A,B,C))

def array_detector_Hermitian(A):
  return np.conj(A.T)

## MF
#@numba.jit(nopython=True)
def array_detector_MF(x,Rinv,S):
  v = array_detector_ABC(array_detector_Hermitian(S),Rinv,x)
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  A = np.linalg.pinv(A)
  t = array_detector_ABC(array_detector_Hermitian(v),A,v)[0,0]
  return np.abs(t)

def MF_pfa(Q , threshold):
    P_fa = 1-stats.chi2.cdf(threshold*2, 2*Q)
    # P_fa = np.exp(-threshold) * sum((threshold**q) / math.factorial(q) for q in range(Q))
    return P_fa

def MF_threshold(Q,pfa):
  return stats.chi2.ppf(1 - pfa, 2 * Q) / 2

def MF_pd(Q , threshold,SCINR_Post):
  # term1 = np.sqrt(2 * SCINR_Post)
  # term2 = np.sqrt(2 * threshold)
  # non_centrality = term1**2
  # Pd = 1 - stats.ncx2.cdf(term2, 2*Q, non_centrality)
  # return Pd
  return 1-stats.ncx2.cdf(x=2 * threshold,df=2*Q,nc=2 * SCINR_Post)
    # P_d_value = sp.marcumq(Q, np.sqrt(2 * SCINR_Post), np.sqrt(2 * threshold))
    # return P_d_value

def array_detector_EstimatedAmplitude(x,Rinv,S):
  v = array_detector_ABC(array_detector_Hermitian(S),Rinv,x)
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  return np.linalg.pinv(A)@v
#@numba.jit(nopython=True)
def array_detector_ED(x,Rinv):
  t = array_detector_ABC(array_detector_Hermitian(x),Rinv,x)[0,0]
  return np.abs(t)

def ED_pfa(Q , threshold):
    P_fa = 1-stats.chi2.cdf(threshold*2, 2*Q)
    # P_fa = np.exp(-threshold) * sum((threshold**q) / math.factorial(q) for q in range(Q))
    return P_fa
#@numba.jit(nopython=True)
def array_detector_ACE(x,Rinv,S):
  return array_detector_MF(x,Rinv,S)/array_detector_ED(x,Rinv)

## Kelly
def array_detector_Kelly(x,Rinv,S):
  return array_detector_MF(x,Rinv,S)/(1+array_detector_ED(x,Rinv))

def Kelly_pfa(Q, NSD, NDim, threshold):
    P_fa = 1 - sp.betainc(Q, NSD - NDim + 1, threshold)
    return P_fa

def Kelly_threshold(Q, NSD, NDim,pfa):
  return sp.betaincinv(Q, NSD - NDim + 1, 1 - pfa)

def Kelly_DistributedIndependentSameRadars_pfa(Q, NSDi, NDimi, threshold):
    P_fa = 0
    for q in range(Q):
      P_fa += (NSDi-NDimi+1)**q / math.factorial(q) * (np.log(threshold)**q)
    P_fa *= threshold ** (-(NSDi-NDimi+1))
    return P_fa

def array_detector_Rao(x,R,S):
  Rinv = np.linalg.pinv(R+array_detector_Hermitian(x)@x)
  v = array_detector_ABC(array_detector_Hermitian(S),Rinv,x)
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  A = np.linalg.pinv(A)
  t = array_detector_ABC(array_detector_Hermitian(v),A,v)[0,0]
  return np.abs(t)

def array_detector_MMED(R):
  eigenvalues = np.linalg.eigvals(R)
  return np.abs(np.max(eigenvalues)/np.min(eigenvalues))

def array_detector_MED(x,R):
  eigenvalues = np.linalg.eigvals(R)
  e = array_detector_Hermitian(x)@x
  return np.abs( e[0,0] / np.min(eigenvalues))

def array_detector_RankOne_SingleShot(X):
  R = X @ array_detector_Hermitian(X)
  eigenvalues = np.linalg.eigvals(R)
  return np.abs( np.max(eigenvalues) / np.trace(R))

def array_detector_SCM_N(X):
  return X @ array_detector_Hermitian(X)

def array_detector_cos2(S,Sp,Rinv):
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  B = array_detector_ABC(array_detector_Hermitian(S),Rinv,Sp)
  BhAinvB = array_detector_ABC(array_detector_Hermitian(B),np.linalg.pinv(A),B)
  C = array_detector_ABC(array_detector_Hermitian(Sp),Rinv,Sp)
  ones = np.ones((S.shape[1],1))
  num = array_detector_Hermitian(ones) @ BhAinvB @ ones
  denum = array_detector_Hermitian(ones) @ C @ ones 
  return np.abs(num[0,0]/denum[0,0])

def array_detector_SINR(S,alpha,Rinv):
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  SINR = array_detector_ABC(array_detector_Hermitian(alpha),A,alpha)
  return np.abs(SINR[0,0])

def array_detector_AMFDeemphasis(x,R,S,eps_AMFD):
  Rinv = np.linalg.pinv(R+eps_AMFD*array_detector_Hermitian(x)@x)
  v = array_detector_ABC(array_detector_Hermitian(S),Rinv,x)
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  A = np.linalg.pinv(A)
  t = array_detector_ABC(array_detector_Hermitian(v),A,v)[0,0]
  return np.abs(t)

def array_detector_Kalson(x,Rinv,S,eps_Kalson):
  return array_detector_MF(x,Rinv,S)/(1+eps_Kalson*array_detector_ED(x,Rinv))

def array_detector_SD(x,Rinv,H): # [50, 51]
  return array_detector_Kelly(x,Rinv,H)

def array_detector_ABORT(x,Rinv,S):
  return (1+array_detector_MF(x,Rinv,S))/(2+array_detector_ED(x,Rinv))

def array_detector_WABORT(x,Rinv,S):
  t = (1-array_detector_Kelly(x,Rinv,S))**2
  A = array_detector_ABC(array_detector_Hermitian(S),Rinv,S)
  ones = np.ones((S.shape[1],1))
  a = np.abs(denum = array_detector_Hermitian(ones) @ A @ ones)[0,0] # Not sure
  return 1 / ((1+a)*t)

def array_detector_CAD(x,Rinv,S,eps_CAD):
  uv = array_detector_ED(x,Rinv) - array_detector_ED(x,Rinv)*(1+eps_CAD**2) # check [41, 58]
  u = 1 if uv >= 0 else 0
  t = array_detector_ED(x,Rinv) - 1/(1-eps_CAD**2)*(np.sqrt(array_detector_ED(x,Rinv)-array_detector_MF(x,Rinv,S))-eps_CAD*np.sqrt(array_detector_MF(x,Rinv,S)))**2 * u  
  return t


def array_detector_CARD(x,Rinv,S,eps_CARD):
  uv = eps_CARD * np.sqrt(array_detector_MF(x,Rinv,S))-np.sqrt(array_detector_ED(x,Rinv)-array_detector_MF(x,Rinv,S))
  u = 1 if uv >= 0 else -1
  return u * uv**2

def array_detector_2SROB(x,Rinv,S,N,K):
  p = array_detector_ED(x,Rinv)-array_detector_MF(x,Rinv,S)
  THR = N/K  # K > 2N
  if p>THR:
    return (THR/p)**THR * np.exp(-THR+array_detector_ED(x,Rinv))
  else:
    return np.exp(array_detector_MF(x,Rinv,S))

def array_detector_1SROB(x,Rinv,S,zeta):
  e = array_detector_ED(x,Rinv)
  p = e - array_detector_MF(x,Rinv,S)
  g = 1 + p
  if p >= 1/(zeta-1):
    Delta = (zeta-1)**(1/zeta) / (1-1/zeta)
    g = (1+Delta) * ( p ** (1/zeta) )
  return (1+e)/g

def array_detector_ROB(x,Rinv,S,zetaeps):
  p = array_detector_ED(x,Rinv)-array_detector_MF(x,Rinv,S)
  xh2 = array_detector_ED(x,Rinv)
  THR = 1/(zetaeps-1)
  if p>THR:
    return ( (1+xh2) * (1-1/zetaeps) ) / ( ((zetaeps-1)*p)**(1/zetaeps) )
  else:
    return (1+xh2)/(1+p)

def hamming(N):
  return scipy_windows.hamming(N)

#@numba.jit(nopython=True)
def array_detector_ROC_MonteCarlo_sortxH0(xH0,xH1,N=200):
    THR_v = xH0   
    THR_v = THR_v[THR_v.argsort()]
    pfa_v = np.logspace(0, -np.log10(THR_v.shape[0]), N)
    indices = []
    Pd_o  = []
    Pfa_o = []
    for pfai in pfa_v:
        ind = round((1 - pfai) * (THR_v.shape[0] - 1))
        if ind not in indices:
            indices.append(ind)
            THR = THR_v[ind]
            pfa = np.sum(xH0>THR)/THR_v.shape[0]
            pd = np.sum(xH1>THR)/THR_v.shape[0]
            Pd_o.append(pd)
            Pfa_o.append(pfa)
    return np.array(Pfa_o),np.array(Pd_o)

#@numba.jit(nopython=True)
def array_detector_ROC_AMF_ACE_MonteCarlo(Min_Pfa,S,SigmaSQRT,NSD,alpha0_SINR,NPfa = 100,progbar = True):
    t_mc = []
    alpha = alpha0_SINR * np.ones((S.shape[1],1))
    signal_H1_0 = S @ alpha
    N_MC_1 = int(NPfa/Min_Pfa)
    mc_100=0
    for mc in range(N_MC_1):
        if progbar:
          if mc_100 < mc * 100 / N_MC_1:
              mc_100+=1
              print(mc_100)
        noiseSD = generate_complex_gaussian_samples(SigmaSQRT,NSD)
        NSigmaHat = noiseSD @ np.conj(noiseSD.T)
        NDim=NSigmaHat.shape[0]
        NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim)
        NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
        N_Trick_MC = 1
        for _ in range(N_Trick_MC):
            noiseH0 = generate_complex_gaussian_samples(SigmaSQRT,1)
            signal_H0 = noiseH0
            t0_ACE = array_detector_ACE(signal_H0,NSigmaHat_DL_inv,S) 
            t0_AMF = array_detector_MF(signal_H0,NSigmaHat_DL_inv,S) 
            noiseH1 = generate_complex_gaussian_samples(SigmaSQRT,1)
            signal_H1 = noiseH1 + signal_H1_0 
            t1_ACE = array_detector_ACE(signal_H1,NSigmaHat_DL_inv,S) 
            t1_AMF = array_detector_MF(signal_H1,NSigmaHat_DL_inv,S) 
            t_mc.append([t0_AMF,t0_ACE,t1_AMF,t1_ACE]) 
    t_mc = np.array(t_mc)
    Pfa_AMF,Pd_AMF=array_detector_ROC_MonteCarlo_sortxH0(t_mc[:, 0],t_mc[:, 2])
    Pfa_ACE,Pd_ACE=array_detector_ROC_MonteCarlo_sortxH0(t_mc[:, 1],t_mc[:, 3])
    return [Pfa_AMF,Pd_AMF],[Pfa_ACE,Pd_ACE]


def array_detector_ROC_Kelly_MonteCarlo(Pfa_v,S,SigmaSQRT,NSD,alpha0_SINR,NMC = [100,100],progbar = True):
  NC1 = NMC[0]
  NC2 = NMC[1]
  iC = Pfa_v * 0
  NC = Pfa_v * 0
  NDim = SigmaSQRT.shape[0]
  for _ in range(NC1):
      noiseSD = generate_complex_gaussian_samples(SigmaSQRT,NSD)
      NSigmaHat = noiseSD @ np.conj(noiseSD.T)
      NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim) 
      NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
      for ics,s in enumerate(Pfa_v):
          thr_Kelly = Kelly_threshold(S.shape[1], NSD, NDim,s)
          alpha = alpha0_SINR * np.ones((S.shape[1],1))
          signal_H1_0 = S @ alpha
          for __ in range(NC2):
              noiseH1 = generate_complex_gaussian_samples(SigmaSQRT,1)
              signal_H1 = noiseH1 + signal_H1_0 
              t1_Kelly = array_detector_Kelly(signal_H1,NSigmaHat_DL_inv,S)
              if t1_Kelly>thr_Kelly:
                  iC[ics]+=1
              NC[ics]+=1
  pd_v = iC / NC
  return pd_v

def array_detector_findTHR_AMF_ACE(Pfa,S,SigmaSQRT,NSD,NPfa = 100,progbar = True):
    t_mc = []
    N_MC_1 = int(NPfa/Pfa)
    mc_100=0
    for mc in range(N_MC_1):
        if progbar:
          if mc_100 < mc * 100 / N_MC_1:
              mc_100+=1
              print(mc_100)
        noiseSD = generate_complex_gaussian_samples(SigmaSQRT,NSD)
        NSigmaHat = noiseSD @ np.conj(noiseSD.T)
        NDim=NSigmaHat.shape[0]
        NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim)
        NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
        N_Trick_MC = 1
        for _ in range(N_Trick_MC):
            noiseH0 = generate_complex_gaussian_samples(SigmaSQRT,1)
            signal_H0 = noiseH0
            t0_ACE = array_detector_ACE(signal_H0,NSigmaHat_DL_inv,S) 
            t0_AMF = array_detector_MF(signal_H0,NSigmaHat_DL_inv,S) 
            t_mc.append([t0_AMF ,t0_ACE]) 

    t_mc = np.array(t_mc)
    N_MC_2 = t_mc.shape[0]
    pfa_v = np.logspace(-np.log10(N_MC_2), 0, 200)
    Pfa_THR_AMF = pfaTHR_tH0(t_mc[:, 0],pfa_v)
    Pfa_THR_ACE = pfaTHR_tH0(t_mc[:, 1],pfa_v)
    thr_AMF = Pfa_THR_AMF[np.argmin(np.abs(Pfa_THR_AMF[:,0]-Pfa)),1]
    thr_ACE = Pfa_THR_ACE[np.argmin(np.abs(Pfa_THR_ACE[:,0]-Pfa)),1]
    return thr_AMF,thr_ACE


def array_detector_PdSNR_AMF_ACE_Kelly_MonteCarlo(Fixed_Pfa,S,SigmaSQRT,NSD,SCINR_Post_v,SCINR_PostProcessingM,NMC = [100,100],progbar = True):
  NC1 = NMC[0]
  NC2 = NMC[1]
  NDim = S.shape[0]
  Q = S.shape[1]
  
  thr_AMF,thr_ACE = array_detector_findTHR_AMF_ACE(Fixed_Pfa,S,SigmaSQRT,NSD) 
  thr_Kelly = Kelly_threshold(Q, NSD, NDim,Fixed_Pfa)
  
  iC = SCINR_Post_v * 0
  iCAMF = SCINR_Post_v * 0
  iCACE = SCINR_Post_v * 0
  NC = SCINR_Post_v * 0
  for isd in range(NC1):
      noiseSD = generate_complex_gaussian_samples(SigmaSQRT,NSD)
      NSigmaHat = noiseSD @ np.conj(noiseSD.T)
      NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim) 
      NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
      for ics,s in enumerate(SCINR_Post_v):
          alpha0 = 10**((s - SCINR_PostProcessingM)/20)
          alpha = alpha0 * np.ones((Q,1))
          signal_H1_0 = S @ alpha
          for isd2 in range(NC2):
              noiseH1 = generate_complex_gaussian_samples(SigmaSQRT,1)
              signal_H1 = noiseH1 + signal_H1_0 
              t1_Kelly = array_detector_Kelly(signal_H1,NSigmaHat_DL_inv,S)
              if t1_Kelly>thr_Kelly:
                  iC[ics]+=1
              t1_Kelly = array_detector_MF(signal_H1,NSigmaHat_DL_inv,S)
              if t1_Kelly>thr_AMF:
                  iCAMF[ics]+=1
              t1_Kelly = array_detector_ACE(signal_H1,NSigmaHat_DL_inv,S)
              if t1_Kelly>thr_ACE:
                  iCACE[ics]+=1
              NC[ics]+=1
  return iCAMF / NC,iCACE / NC,iC / NC
  
def array_detector_PdNSecondary_AMF_ACE_Kelly_MonteCarlo(Fixed_Pfa,S,SigmaSQRT,NSDTimes_v,SCINR_Post_Fixed,SCINR_PostProcessingM,NMC = [100,100],progbar = True):
  NC1 = NMC[0]
  NC2 = NMC[1]
  NDim = S.shape[0]
  Q = S.shape[1]
  thr_AMF=[]
  thr_ACE=[]
  thr_Kelly=[]
  for NSDTimes in NSDTimes_v:
    NSD = int(NSDTimes*NDim)
    thr_AMF0,thr_ACE0 = array_detector_findTHR_AMF_ACE(Fixed_Pfa,S,SigmaSQRT,NSD,NC1) 
    thr_Kelly0 = Kelly_threshold(Q, NSD, NDim,Fixed_Pfa)
    thr_AMF.append(thr_AMF0)
    thr_ACE.append(thr_ACE0)
    thr_Kelly.append(thr_Kelly0)
    
  iC = NSDTimes_v * 0
  iCAMF = NSDTimes_v * 0
  iCACE = NSDTimes_v * 0
  NC = NSDTimes_v * 0
  for ics,s in enumerate(NSDTimes_v):
      NSD =int(s*NDim)
      for isd in range(NC1):
        noiseSD = generate_complex_gaussian_samples(SigmaSQRT,NSD)
        NSigmaHat = noiseSD @ np.conj(noiseSD.T)
        NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim) 
        NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
        alpha0 = 10**((SCINR_Post_Fixed - SCINR_PostProcessingM)/20)
        alpha = alpha0 * np.ones((Q,1))
        signal_H1_0 = S @ alpha
        for isd2 in range(NC2):
            noiseH1 = generate_complex_gaussian_samples(SigmaSQRT,1)
            signal_H1 = noiseH1 + signal_H1_0 
            t1_Kelly = array_detector_Kelly(signal_H1,NSigmaHat_DL_inv,S)
            if t1_Kelly>thr_Kelly[ics]:
                iC[ics]+=1
            t1_Kelly = array_detector_MF(signal_H1,NSigmaHat_DL_inv,S)
            if t1_Kelly>thr_AMF[ics]:
                iCAMF[ics]+=1
            t1_Kelly = array_detector_ACE(signal_H1,NSigmaHat_DL_inv,S)
            if t1_Kelly>thr_ACE[ics]:
                iCACE[ics]+=1
            NC[ics]+=1
  return iCAMF / NC,iCACE / NC,iC / NC
  
def VAsignal2coArraySignal_1D_RangeDopplerVA(virtual_array_Position, RDvirtual_array_Signal):
    coprime_differences = []  # To store unique element differences
    coprime_struct = {}  # Dictionary to store co-prime pairs
    
    # Calculate all pairwise differences for the co-prime array
    for i in range(len(virtual_array_Position)):
        for j in range(len(virtual_array_Position)):
            difference = virtual_array_Position[i] - virtual_array_Position[j]
            if difference not in coprime_differences:
                coprime_differences.append(difference)
                coprime_struct[difference] = [(i, j)]
            else:
                coprime_struct[difference].append((i, j))
    
    coprime_differences = np.array(sorted(coprime_differences))
    sorted_indices = np.argsort(coprime_differences)
    
    coArraySignal = np.zeros((RDvirtual_array_Signal.shape[0],RDvirtual_array_Signal.shape[1],len(coprime_differences)), dtype=complex)
    # co_prime_steering_vectors = np.zeros((len(coprime_differences), len(sin_angles)), dtype=complex)
    
    for i, diff in enumerate(coprime_differences):
        pairs = coprime_struct[diff]
        
        # Average over all pairs for each co-prime difference
        for idx1, idx2 in pairs:
            coArraySignal[:,:,i] += RDvirtual_array_Signal[:,:,idx1] * np.conj(RDvirtual_array_Signal[:,:,idx2])
        coArraySignal[:,:,i] /= len(pairs)  # Average
        
        # Calculate the steering vector for each co-prime difference
        # co_prime_steering_vectors[i, :] = np.exp(1j * np.pi * diff * sin_angles)
    
    coArrayStructure = [coprime_struct,coprime_differences]
    return coArraySignal,coArrayStructure
def VAsignal2coArraySignal_1D(virtual_array_Position, virtual_array_Signal):
    coprime_differences = []  # To store unique element differences
    coprime_struct = {}  # Dictionary to store co-prime pairs
    
    # Calculate all pairwise differences for the co-prime array
    for i in range(len(virtual_array_Position)):
        for j in range(len(virtual_array_Position)):
            difference = virtual_array_Position[i] - virtual_array_Position[j]
            if difference not in coprime_differences:
                coprime_differences.append(difference)
                coprime_struct[difference] = [(i, j)]
            else:
                coprime_struct[difference].append((i, j))
    
    coprime_differences = np.array(sorted(coprime_differences))
    sorted_indices = np.argsort(coprime_differences)
    
    coArraySignal = np.zeros(len(coprime_differences), dtype=complex)
    # co_prime_steering_vectors = np.zeros((len(coprime_differences), len(sin_angles)), dtype=complex)
    
    for i, diff in enumerate(coprime_differences):
        pairs = coprime_struct[diff]
        
        # Average over all pairs for each co-prime difference
        for idx1, idx2 in pairs:
            coArraySignal[i] += virtual_array_Signal[idx1] * np.conj(virtual_array_Signal[idx2])
        coArraySignal[i] /= len(pairs)  # Average
        
        # Calculate the steering vector for each co-prime difference
        # co_prime_steering_vectors[i, :] = np.exp(1j * np.pi * diff * sin_angles)
    
    coArrayStructure = [coprime_struct,coprime_differences]
    return coArraySignal,coArrayStructure
def VAsignal2coArraySignal_1D_sum(virtual_array_Position, virtual_array_Signal):
    coprime_differences = []  # To store unique element differences
    coprime_struct = {}  # Dictionary to store co-prime pairs
    
    # Calculate all pairwise differences for the co-prime array
    for i in range(len(virtual_array_Position)):
        for j in range(len(virtual_array_Position)):
            difference = virtual_array_Position[i] - virtual_array_Position[j]
            if difference not in coprime_differences:
                coprime_differences.append(difference)
                coprime_struct[difference] = [(i, j,-1)]
            else:
                coprime_struct[difference].append((i, j,-1))
            difference = virtual_array_Position[i] + virtual_array_Position[j]
            if difference not in coprime_differences:
                coprime_differences.append(difference)
                coprime_struct[difference] = [(i, j,1)]
            else:
                coprime_struct[difference].append((i, j,1))
    
    coprime_differences = np.array(sorted(coprime_differences))
    sorted_indices = np.argsort(coprime_differences)
    
    coArraySignal = np.zeros(len(coprime_differences), dtype=complex)
    # co_prime_steering_vectors = np.zeros((len(coprime_differences), len(sin_angles)), dtype=complex)
    
    for i, diff in enumerate(coprime_differences):
        pairs = coprime_struct[diff]
        
        # Average over all pairs for each co-prime difference
        for idx1, idx2,sd in pairs:
            if sd == -1:
              coArraySignal[i] += virtual_array_Signal[idx1] * np.conj(virtual_array_Signal[idx2])
            else:
              coArraySignal[i] += virtual_array_Signal[idx1] * (virtual_array_Signal[idx2])
        coArraySignal[i] /= len(pairs)  # Average
        
        # Calculate the steering vector for each co-prime difference
        # co_prime_steering_vectors[i, :] = np.exp(1j * np.pi * diff * sin_angles)
    
    coArrayStructure = [coprime_struct,coprime_differences]
    return coArraySignal,coArrayStructure



import bpy
# from .. import constants
from ..constants import *
from ..environment import BlenderSuiteFinder
from .BlenderGeometry import *
from .RayTracingFunctions import *
import sensingsp as ssp
import math
class Vector3D:
    def __init__(self, x, y, z):
        self.x = np.longdouble(x)
        self.y = np.longdouble(y)
        self.z = np.longdouble(z)

    def distance_to(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        return np.sqrt(dx**2 + dy**2 + dz**2)

    def diff(self, other):
        return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)

    def __repr__(self):
        return f"Vector3D(x={self.x}, y={self.y}, z={self.z})"
# x = Vector3D(1,2,3)
# print(x)
def Path_RayTracing(progressbar=[]):
    all_info = []
    all_renderImages = []
    all_info_TXRXPos = []
    ssp.lastScatterInfo = []
    
    geocalculator = BlenderGeometry()
    rayTracingFunctions = RayTracingFunctions()


    if "Simulation Settings" in bpy.data.objects:
        sim_axes = bpy.data.objects["Simulation Settings"]

        if "Velocity Calc. Jump" not in sim_axes:
          sim_axes["Velocity Calc. Jump"]=1
        blender_frame_Jump_for_Velocity = sim_axes["Velocity Calc. Jump"]

        BounceNumber = bpy.data.objects["Simulation Settings"]["Bounce Number"]
        RenderBlenderFrames = bpy.data.objects["Simulation Settings"]["Render Blender Frames"]
        render_imade_dir = bpy.data.objects["Simulation Settings"]["Video Directory"]
        open_output_folder = bpy.data.objects["Simulation Settings"]["Open Output Folder"]

    else:
        blender_frame_Jump_for_Velocity = 1
        BounceNumber = 2
        RenderBlenderFrames = False
        render_imade_dir = current_working_directory
        open_output_folder = True

    scene = bpy.context.scene
    suite_information = BlenderSuiteFinder().find_suite_information()

    anything2do = True
    blender_frame_index = scene.frame_start
    while anything2do:
        print(f'frame {blender_frame_index}')
        if blender_frame_index+blender_frame_Jump_for_Velocity>scene.frame_end:
            break
        if not(progressbar==[]):
            progressbar.progress_update(0+50*blender_frame_index/scene.frame_end)
        anything2do = False

        scene.frame_set(blender_frame_index)
        bpy.context.view_layer.update()
        depsgraph = bpy.context.evaluated_depsgraph_get()

        Suite_Position,ScattersGeo,HashFaceIndex_ScattersGeo,ScattersGeoV = geocalculator.get_Position_Velocity(scene, suite_information, blender_frame_index, blender_frame_Jump_for_Velocity)

        ssp.lastScatterInfo.append(ScattersGeo)
        print(f"ScattersGeo {len(ScattersGeo)}")
        # plotscene(Suite_Position,ScattersGeo)

        ## RIS init
        for isuite,suiteobject in enumerate(suite_information):
          if 1: # Zero Phase
            for iris , ris in enumerate(Suite_Position[isuite]['RIS']):
                for iriselement ,ris_element_position in enumerate(ris['Position']):
                    ris['PhaseAmplitude'][iriselement][0] = 1
                    ris['PhaseAmplitude'][iriselement][1] = 0
          else: # Random Phase
            for iris , ris in enumerate(Suite_Position[isuite]['RIS']):
                for iriselement ,ris_element_position in enumerate(ris['Position']):
                    ris['PhaseAmplitude'][iriselement][0] = 1
                    ris['PhaseAmplitude'][iriselement][1] = 2*np.pi*np.random.rand()


        # Probes Reset
        for isuite,suiteobject in enumerate(suite_information):
          for iprobe , probe in enumerate(Suite_Position[isuite]['Probe']):
            for iprobe2 , probe2 in enumerate(probe):
              for iprobev , probev in enumerate(probe2):
                probev[2]=np.longdouble(0)+1j*np.longdouble(0)

        all_d_drate_amp = {}
        for isuite in range(len(suite_information)):
            all_d_drate_amp[isuite] = {}
            for iradar in range(len(suiteobject['Radar'])):
                all_d_drate_amp[isuite][iradar] = {}
                for irx in range(len(Suite_Position[isuite]['Radar'][iradar]['RX-Position'])):
                    all_d_drate_amp[isuite][iradar][irx] = {}
                    for isuite2 in range(len(suite_information)):
                        all_d_drate_amp[isuite][iradar][irx][isuite2] = {}
                        for iradar2 in range(len(suiteobject['Radar'])):
                            all_d_drate_amp[isuite][iradar][irx][isuite2][iradar2] = {}
                            for itx in range(len(Suite_Position[isuite2]['Radar'][iradar2]['TX-Position'])):
                                all_d_drate_amp[isuite][iradar][irx][isuite2][iradar2][itx] = []


        for isuite,suiteobject in enumerate(suite_information):
          ## Radar Analysis
          for iradar,radarobject in enumerate(suiteobject['Radar']):
            # Timing should be done
            print(isuite,iradar)
            raysets = []
            for itx in range(len(Suite_Position[isuite]['Radar'][iradar]['TX-Position'])):
              rayset = rayTracingFunctions.rayset_gen_TX(itx,Suite_Position,isuite,iradar,ScattersGeo,depsgraph)
              raysets.append(rayset)

            Paths = []
            # BounceNumber_Radar = 2
            # BounceNumber_RIS = 2
            # bounce = 0
            # while 1:
            #   if

            for bounce in range(BounceNumber):

              raysetsNext = []
              for rayset in raysets:
                rxPath = rayTracingFunctions.RXPath(rayset)
                for path in rxPath:
                  Paths.append(path)
                for _ in rayset.metaInformation['RIS']:
                  rayset_new = rayTracingFunctions.rayset_gen_RIS(_[2],Suite_Position,_[0],_[1],ScattersGeo,depsgraph)
                  rayset_new.source_ray = rayset
                  raysetsNext.append(rayset_new)
                for targetLOS_hitpoint_startreflectedpoints_randomvectors in rayset.metaInformation['Target']:
                  rayset_new = rayTracingFunctions.rayset_gen_Scatter(targetLOS_hitpoint_startreflectedpoints_randomvectors,Suite_Position,isuite,iradar,depsgraph)
                  rayset_new.source_ray = rayset
                  raysetsNext.append(rayset_new)
              raysets = raysetsNext

            for path in Paths:
              amp = 1
              d = np.longdouble(0)
              drate = 0
              # DESIRED FOR EACH PATH: AMP D DRATE

              # if len(path.middle_elements)==0: # Strong Interference + Friis Equation
              #   continue

              isuiteTX = path.transmitter[0]
              iradarTX = path.transmitter[1]
              itx = path.transmitter[2]
              tx = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Position'][itx]
              tx_v = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Velocity'][itx]
              TX = Vector3D(tx.x,tx.y,tx.z)
              Scatters_in_the_Middle = []
              if len(path.middle_elements)==0:

                isuiteRX = path.receiver[0]
                iradarRX = path.receiver[1]
                irx = path.receiver[2]
                next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
                Next = Vector3D(next.x,next.y,next.z)
                nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]

                dir = next-tx
                distanceLP =  dir.magnitude
                dir/= distanceLP
                # Distance
                d+=Next.distance_to(TX)
                # DistanceRate
                drate += (-tx_v.dot(dir)+nextv.dot(dir))
                # Amplitude 1
                if Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx]=='Omni':
                  gain = 1
                else:
                  az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Direction'][itx],dir)
                  gain = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
                  gain*= np.sinc(az/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx])*np.sinc(el/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx])

                Total_Transmit_Power = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Power'][itx]
                PTX = gain * Total_Transmit_Power
                amp*=np.sqrt(PTX)/distanceLP/np.sqrt(4*np.pi)

                # Amplitude 2
                if Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx]=='Omni':
                  gain = 1
                else:
                  az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-'][irx],dir)
                  gain = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternMaxGain'][irx]
                  gain*= np.sinc(az/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx])*np.sinc(el/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx])
                Lambda = Suite_Position[isuiteRX]['Radar'][iradarRX]['WaveLength']
                Aeff = gain * Lambda * Lambda / (4 * np.pi)
                amp*=np.sqrt(Aeff)
                all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append([d,drate,amp,Scatters_in_the_Middle])
                continue
              # find Next for TX (Step 0)
              next_element = path.middle_elements[len(path.middle_elements)-1]
              if next_element.source_type=='scatter':
                next = next_element.ids[1]
                nextv = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
                next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
                Next = Vector3D(next.x,next.y,next.z)
                Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])
              if next_element.source_type=='RIS':
                isuiteRIS = next_element.ids[0]
                iris = next_element.ids[1]
                iriselement = next_element.ids[2]
                next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
                Next = Vector3D(next.x,next.y,next.z)
                nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
              dir = next-tx
              distanceLP =  dir.magnitude
              dir/= distanceLP
              # Distance
              d+=Next.distance_to(TX)
              # DistanceRate
              drate += (-tx_v.dot(dir)+nextv.dot(dir))
              # Amplitude 1
              if Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx]=='Omni':
                gain = 1
              else:
                az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Direction'][itx],dir)
                gain = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
                gain*= np.sinc(az/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx])*np.sinc(el/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx])

              Total_Transmit_Power = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Power'][itx]
              PTX = gain * Total_Transmit_Power
              amp*=np.sqrt(PTX)
              amp/=distanceLP*np.sqrt(4*np.pi)
              # print(amp,"1,1",PTX,distanceLP,"np.sqrt(PTX)/distanceLP/np.sqrt(4*np.pi)")
              # Amplitude 2
              if next_element.source_type=='scatter':
                area = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
                RCS0 = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
                area*=np.abs(next_normal.dot(-dir))
                rcs = np.sqrt(area * RCS0)
                amp*=np.sqrt(rcs)
              # print(amp,"1,2",rcs,"RCS")
              if next_element.source_type=='RIS':
                isuiteRIS = next_element.ids[0]
                iris = next_element.ids[1]
                iriselement = next_element.ids[2]
                if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
                  gain = 1
                else:
                  az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
                  gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
                  gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

                Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
                Aeff = gain * Lambda * Lambda / (4 * np.pi)
                amp*=np.sqrt(Aeff)
              # Middle Elements
              for imiddle_elements in range(len(path.middle_elements)-1):
                middle_element = path.middle_elements[len(path.middle_elements)-1-imiddle_elements]
                if middle_element.source_type=='scatter':
                  p = middle_element.ids[1]
                  currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
                  currentPoint = Vector3D(p.x,p.y,p.z)
                  normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
                  area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
                  RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
                  Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]],currentPoint])
                  next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
                  if next_element.source_type=='scatter':
                    next = next_element.ids[1]
                    nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
                    next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
                    Next = Vector3D(next.x,next.y,next.z)
                    dir = next-p
                    distanceLP =  dir.magnitude
                    dir/= distanceLP
                    # Distance
                    d+=Next.distance_to(currentPoint)
                    # DistanceRate
                    drate += (-currentv.dot(dir)+nextv.dot(dir))
                    # Amplitude 1
                    area*=np.abs(normal.dot(dir))
                    rcs = np.sqrt(area * RCS0)
                    amp*=np.sqrt(rcs)
                    amp/=distanceLP*np.sqrt(4*np.pi)
                    # Amplitude 2
                    area     = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
                    area*=np.abs(next_normal.dot(-dir))
                    RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
                    rcs = np.sqrt(area * RCS0)
                    amp*=np.sqrt(rcs)

                  if next_element.source_type=='RIS':
                    isuiteRIS = next_element.ids[0]
                    iris = next_element.ids[1]
                    iriselement = next_element.ids[2]
                    next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
                    Next = Vector3D(next.x,next.y,next.z)
                    nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
                    dir = next-p
                    distanceLP =  dir.magnitude
                    dir/= distanceLP
                    # Distance
                    d+=Next.distance_to(currentPoint)
                    # DistanceRate
                    drate += (-currentv.dot(dir)+nextv.dot(dir))
                    # Amplitude 1
                    area*=np.abs(normal.dot(dir))
                    rcs = np.sqrt(area * RCS0)
                    amp*=np.sqrt(rcs)
                    amp/=distanceLP*np.sqrt(4*np.pi)
                    # Amplitude 2
                    if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
                      gain = 1
                    else:
                      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
                      gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
                      gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

                    Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
                    Aeff = gain * Lambda * Lambda / (4 * np.pi)
                    amp*=np.sqrt(Aeff)
                if middle_element.source_type=='RIS':
                  isuiteRIS0 = middle_element.ids[0]
                  iris0 = middle_element.ids[1]
                  iriselement0 = middle_element.ids[2]
                  p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
                  currentPoint = Vector3D(p.x,p.y,p.z)
                  currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
                  next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
                  if next_element.source_type=='scatter':
                    next = next_element.ids[1]
                    nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
                    next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
                    Next = Vector3D(next.x,next.y,next.z)
                    dir = next-p
                    distanceLP =  dir.magnitude
                    dir/= distanceLP
                    # Distance
                    d+=Next.distance_to(currentPoint)
                    # DistanceRate
                    drate += (-currentv.dot(dir)+nextv.dot(dir))
                    # Amplitude 1
                    if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
                      gain = 1
                    else:
                      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
                      gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
                      gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
                    amp*=np.sqrt(gain)
                    amp/=distanceLP*np.sqrt(4*np.pi)
                    # Amplitude 2
                    area     = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
                    area*=np.abs(next_normal.dot(-dir))
                    RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
                    rcs = np.sqrt(area * RCS0)
                    amp*=np.sqrt(rcs)

                  if next_element.source_type=='RIS':
                    isuiteRIS = next_element.ids[0]
                    iris = next_element.ids[1]
                    iriselement = next_element.ids[2]
                    next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
                    Next = Vector3D(next.x,next.y,next.z)
                    nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
                    dir = next-p
                    distanceLP =  dir.magnitude
                    dir/= distanceLP
                    # Distance
                    d+=Next.distance_to(currentPoint)
                    # DistanceRate
                    drate += (-currentv.dot(dir)+nextv.dot(dir))
                    # Amplitude 1
                    if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
                      gain = 1
                    else:
                      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
                      gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
                      gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
                    amp*=np.sqrt(gain)
                    amp/=distanceLP*np.sqrt(4*np.pi)
                    # Amplitude 2
                    if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
                      gain = 1
                    else:
                      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
                      gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
                      gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

                    Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
                    Aeff = gain * Lambda * Lambda / (4 * np.pi)
                    amp*=np.sqrt(Aeff)

              # Last
              middle_element = path.middle_elements[0]
              isuiteRX = path.receiver[0]
              iradarRX = path.receiver[1]
              irx = path.receiver[2]
              next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
              Next = Vector3D(next.x,next.y,next.z)
              nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]
              if middle_element.source_type=='scatter':
                p = middle_element.ids[1]
                currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
                currentPoint = Vector3D(p.x,p.y,p.z)
                normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
                area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
                RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
              if middle_element.source_type=='RIS':
                isuiteRIS0 = middle_element.ids[0]
                iris0 = middle_element.ids[1]
                iriselement0 = middle_element.ids[2]
                p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
                currentPoint = Vector3D(p.x,p.y,p.z)
                currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
                # next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
              dir = next-p
              distanceLP =  dir.magnitude
              dir/= distanceLP
              # Distance
              d+=Next.distance_to(currentPoint)
              # DistanceRate
              drate += (-currentv.dot(dir)+nextv.dot(dir))
              # Amplitude 1
              if middle_element.source_type=='scatter':
                area*=np.abs(normal.dot(dir))
                rcs = np.sqrt(area * RCS0)
                amp*=np.sqrt(rcs)
              # print(amp,"1,3",rcs,"RCS")
              if middle_element.source_type=='RIS':
                if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
                  gain = 1
                else:
                  az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
                  gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
                  gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
                amp*=np.sqrt(gain)

              amp/=distanceLP*np.sqrt(4*np.pi)
              # Amplitude 2
              if Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx]=='Omni':
                gain = 1
              else:
                az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-'][irx],dir)
                gain = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternMaxGain'][irx]
                gain*= np.sinc(az/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx])*np.sinc(el/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx])
              Lambda = Suite_Position[isuiteRX]['Radar'][iradarRX]['WaveLength']
              Aeff = gain * Lambda * Lambda / (4 * np.pi)
              amp*=np.sqrt(Aeff)
              # print(amp,"1,4",Aeff,gain,Lambda,"gain * Lambda * Lambda / (4 * np.pi)")
              all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append([d,drate,amp,Scatters_in_the_Middle])


        # print(all_d_drate_amp)
        all_info.append(all_d_drate_amp)
        TXRXPos=[]
        for isuite in range(len(suite_information)):
            RXTX=[]
            for iradar in range(len(suiteobject['Radar'])):
                RXTX.append([Suite_Position[isuite]['Radar'][iradar]['TX-Position'],Suite_Position[isuite]['Radar'][iradar]['RX-Position']])
            TXRXPos.append(RXTX)
        all_info_TXRXPos.append(TXRXPos)
        # if RenderBlenderFrames:
        #   suiteims = []
        #   for isuite,suiteobject in enumerate(suite_information):
        #     radims=[]
        #     for iradar,radarobject in enumerate(suiteobject['Radar']):
        #       if len(radarobject['RX'])>0:
        #         bpy.context.scene.camera = radarobject['RX'][0]
        #         bpy.context.scene.render.resolution_x = 640
        #         bpy.context.scene.render.resolution_y = 480
        #         bpy.context.scene.render.resolution_percentage = 100
        #         bpy.context.scene.render.filepath = f"{render_imade_dir}/rendered_image.png"
        #         bpy.ops.render.render(write_still=True)
        #         image = Image.open(bpy.context.scene.render.filepath)
        #         image_np = np.array(image)
        #         radims.append(image_np)
        #     suiteims.append(radims)
        #   all_renderImages.append(suiteims)

        blender_frame_index+=1
        anything2do = True
        # break
    ssp.Paths = all_info
def Path_RayTracing_frame(progressbar=[]):
    empty = bpy.data.objects.get("Detection Cloud")
    if empty:
        # Select all children of the empty object
        for child in empty.children:
            child.select_set(True)
        # Select the empty itself
        empty.select_set(True)
        # Delete selected objects
        bpy.ops.object.delete()
    
    bpy.context.scene.frame_set(ssp.config.CurrentFrame)
    bpy.context.view_layer.update()
    depsgraph = bpy.context.evaluated_depsgraph_get()
    geocalculator = BlenderGeometry()
    rayTracingFunctions = RayTracingFunctions()
    suite_information = BlenderSuiteFinder().find_suite_information()
    
    if "Simulation Settings" in bpy.data.objects:
        sim_axes = bpy.data.objects["Simulation Settings"]
        if "Velocity Calc. Jump" not in sim_axes:
          sim_axes["Velocity Calc. Jump"]=1
        blender_frame_Jump_for_Velocity = sim_axes["Velocity Calc. Jump"]
        BounceNumber = bpy.data.objects["Simulation Settings"]["Bounce Number"]
        # RenderBlenderFrames = bpy.data.objects["Simulation Settings"]["Render Blender Frames"]
        # render_imade_dir = bpy.data.objects["Simulation Settings"]["Video Directory"]
        do_RayTracing_LOS = bpy.data.objects["Simulation Settings"]["do RayTracing LOS"]
        do_RayTracing_Simple = bpy.data.objects["Simulation Settings"]["do RayTracing Simple"]
    else:
        blender_frame_Jump_for_Velocity = 1
        BounceNumber = 2
        do_RayTracing_LOS = True 
        do_RayTracing_Simple = False
        # RenderBlenderFrames = False
        # render_imade_dir = current_working_directory
    Suite_Position,ScattersGeo,HashFaceIndex_ScattersGeo,ScattersGeoV = geocalculator.get_Position_Velocity(bpy.context.scene, suite_information, ssp.config.CurrentFrame, blender_frame_Jump_for_Velocity)
    # float(bpy.context.scene.render.fps)
    ssp.lastScatterInfo = ScattersGeo
    ssp.lastSuite_Position=Suite_Position
    ## RIS init
    # for isuite,suiteobject in enumerate(suite_information):
    #   if 1: # Zero Phase
    #     for iris , ris in enumerate(Suite_Position[isuite]['RIS']):
    #         for iriselement ,ris_element_position in enumerate(ris['Position']):
    #             ris['PhaseAmplitude'][iriselement][0] = 1
    #             ris['PhaseAmplitude'][iriselement][1] = 0
    #   else: # Random Phase
    #     for iris , ris in enumerate(Suite_Position[isuite]['RIS']):
    #         for iriselement ,ris_element_position in enumerate(ris['Position']):
    #             ris['PhaseAmplitude'][iriselement][0] = 1
    #             ris['PhaseAmplitude'][iriselement][1] = 2*np.pi*np.random.rand()
    # Probes Reset
    for isuite,suiteobject in enumerate(suite_information):
      for iprobe , probe in enumerate(Suite_Position[isuite]['Probe']):
        for iprobe2 , probe2 in enumerate(probe):
          for iprobev , probev in enumerate(probe2):
            probev[2]=np.longdouble(0)+1j*np.longdouble(0)
    all_d_drate_amp = {}
    for isuite in range(len(suite_information)):
        all_d_drate_amp[isuite] = {}
        for iradar in range(len(suiteobject['Radar'])):
            all_d_drate_amp[isuite][iradar] = {}
            for irx in range(len(Suite_Position[isuite]['Radar'][iradar]['RX-Position'])):
                all_d_drate_amp[isuite][iradar][irx] = {}
                for isuite2 in range(len(suite_information)):
                    all_d_drate_amp[isuite][iradar][irx][isuite2] = {}
                    for iradar2 in range(len(suiteobject['Radar'])):
                        all_d_drate_amp[isuite][iradar][irx][isuite2][iradar2] = {}
                        for itx in range(len(Suite_Position[isuite2]['Radar'][iradar2]['TX-Position'])):
                            all_d_drate_amp[isuite][iradar][irx][isuite2][iradar2][itx] = []
    for isuite,suiteobject in enumerate(suite_information):
      ## Radar Analysis
      for iradar,radarobject in enumerate(suiteobject['Radar']):
        # Timing should be done
        # print(isuite,iradar)
        Paths = []
        if do_RayTracing_LOS:
          if do_RayTracing_Simple:
            rad = Suite_Position[isuite]['Radar'][iradar]['TX-Position']
            cent = Vector((0,0,0))
            for pos in rad:
              cent+=pos
            cent/=len(rad)
            LOS_target = []
            for target in ScattersGeo:
              if rayTracingFunctions.check_line_of_sight(cent, target[0], depsgraph):
                LOS_target.append(target)
            for itx in range(len(Suite_Position[isuite]['Radar'][iradar]['TX-Position'])):
              for target in LOS_target: #face_center_all.append([face_center,obj_hash,face.index,face.calc_area(),fn,RCS0])
                for irx in range(len(Suite_Position[isuite]['Radar'][iradar]['RX-Position'])):
                    path = Path4WavePropagation(None,None,[])
                    path.transmitter = [isuite,iradar,itx]
                    mid=SourceType("scatter", [[-1,target[1],target[2]],target[0]])
                    path.middle_elements.append(mid)
                    path.receiver = [isuite,iradar,irx]
                    Paths.append(path)
          else:
            raysets = []
            for itx in range(len(Suite_Position[isuite]['Radar'][iradar]['TX-Position'])):
              rayset = rayTracingFunctions.rayset_gen_TX(itx,Suite_Position,isuite,iradar,ScattersGeo,depsgraph)
              raysets.append(rayset)
            for bounce in range(BounceNumber):
              raysetsNext = []
              for rayset in raysets:
                rxPath = rayTracingFunctions.RXPath(rayset)
                for path in rxPath:
                  Paths.append(path)
                for _ in rayset.metaInformation['RIS']:
                  rayset_new = rayTracingFunctions.rayset_gen_RIS(_[2],Suite_Position,_[0],_[1],ScattersGeo,depsgraph)
                  rayset_new.source_ray = rayset
                  raysetsNext.append(rayset_new)
                # Test = len(rayset.metaInformation['Target'])
                # print("Test ",Test)
                for targetLOS_hitpoint_startreflectedpoints_randomvectors in rayset.metaInformation['Target']:
                  rayset_new = rayTracingFunctions.rayset_gen_Scatter(targetLOS_hitpoint_startreflectedpoints_randomvectors,Suite_Position,isuite,iradar,depsgraph)
                  if len(rayset_new.metaInformation['RX'])+len(rayset_new.metaInformation['Target'])+len(rayset_new.metaInformation['RIS'])>0:
                    rayset_new.source_ray = rayset
                    raysetsNext.append(rayset_new)
              raysets = raysetsNext
        else:
          # import sys
          size=0
          for itx in range(len(Suite_Position[isuite]['Radar'][iradar]['TX-Position'])):
            for itarget,target in enumerate(ScattersGeo): #face_center_all.append([face_center,obj_hash,face.index,face.calc_area(),fn,RCS0])
              for irx in range(len(Suite_Position[isuite]['Radar'][iradar]['RX-Position'])):
                  path = Path4WavePropagation(None,None,[])
                  path.transmitter = [isuite,iradar,itx]
                  mid=SourceType("scatter", [[-1,target[1],target[2]],target[0]])
                  path.middle_elements.append(mid)
                  path.receiver = [isuite,iradar,irx]
                  Paths.append(path)
          #         size += sys.getsizeof(path)
          # size2 = sys.getsizeof(Paths)

        if 1:
          paths_processing(Paths,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,all_d_drate_amp)
        else:
          for path in Paths:
            amp = np.longdouble(1)
            d = np.longdouble(0)
            drate = 0
            # DESIRED FOR EACH PATH: AMP D DRATE

            # if len(path.middle_elements)==0: # Strong Interference + Friis Equation
            #   continue
            d_drate_amp_Scatters_in_the_Middle,is_end=calc_Path_TX(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp)
            if is_end:
              RXfromTX = ssp.config.directReceivefromTX
              if RXfromTX:
                all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append(d_drate_amp_Scatters_in_the_Middle)
              continue
            d,drate,amp,Scatters_in_the_Middle = d_drate_amp_Scatters_in_the_Middle
            
            # Middle Elements
            d,drate,amp,Scatters_in_the_Middle=calc_Path_Middle(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle)
            
            d,drate,amp=calc_Path_RX(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp)
            
            all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append([d,drate,amp,Scatters_in_the_Middle])
    return all_d_drate_amp
def calc_Path_TX(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp):
  isuiteTX = path.transmitter[0]
  iradarTX = path.transmitter[1]
  itx = path.transmitter[2]
  tx = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Position'][itx]
  tx_v = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Velocity'][itx]
  TX = Vector3D(tx.x,tx.y,tx.z)
  Scatters_in_the_Middle = []
  if len(path.middle_elements)==0:

    isuiteRX = path.receiver[0]
    iradarRX = path.receiver[1]
    irx = path.receiver[2]
    next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
    Next = Vector3D(next.x,next.y,next.z)
    nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]
    dir = next-tx
    distanceLP =  dir.magnitude
    dir/= distanceLP
    # Distance
    d+=Next.distance_to(TX)
    # DistanceRate
    drate += (-tx_v.dot(dir)+nextv.dot(dir))
    # Amplitude 1
    if Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx]=='Omni':
      gain = 1
    else:
      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Direction'][itx],dir)
      gain = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
      # gain = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
    
      gain*=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(az), np.rad2deg(el), 
                                                  1, Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx],
                                                    Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx],
                                                    Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx])
      
          
      # gain*= np.sinc(az/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx])*np.sinc(el/Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx])

    Total_Transmit_Power = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Power'][itx]
    PTX = gain * Total_Transmit_Power
    amp*=np.sqrt(PTX)/distanceLP/np.sqrt(4*np.pi)

    # Amplitude 2
    if Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx]=='Omni':
      gain = 1
    else:
      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Direction'][irx],-dir)
      gain = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternMaxGain'][irx]
      gain*=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(az), np.rad2deg(el), 
                                                1, Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx],
                                                  Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx],
                                                  Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx])
      # gain*= np.sinc(az/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx])*np.sinc(el/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx])
    Lambda = Suite_Position[isuiteRX]['Radar'][iradarRX]['WaveLength']
    Aeff = gain * Lambda * Lambda / (4 * np.pi)
    amp*=np.sqrt(Aeff)
    # all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append([d,drate,amp,Scatters_in_the_Middle])
    return [d,drate,amp,Scatters_in_the_Middle],True
  # find Next for TX (Step 0)
  next_element = path.middle_elements[len(path.middle_elements)-1]
  if next_element.source_type=='scatter':
    next = next_element.ids[1]
    nextv = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
    next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
    Next = Vector3D(next.x,next.y,next.z)
    Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])
  if next_element.source_type=='RIS':
    isuiteRIS = next_element.ids[0]
    iris = next_element.ids[1]
    iriselement = next_element.ids[2]
    next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
    Next = Vector3D(next.x,next.y,next.z)
    nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
    
    risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
    risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
    risd *= (math.fmod(risphase,360)/360)/2
    amp*=np.sqrt(np.sqrt(risamp))
    d+=risd
    # print(next_element.ids,next,nextv)
    
  dir = next-tx
  distanceLP =  dir.magnitude
  dir/= distanceLP
  # Distance
  d+=Next.distance_to(TX)
  # DistanceRate
  drate += (-tx_v.dot(dir)+nextv.dot(dir))
  # Amplitude 1
  if Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx]=='Omni':
    gain = 1
  else:
    az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Direction'][itx],dir)
    gain = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
    
    gain*=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(az), np.rad2deg(el), 
                                                1, Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx],
                                                  Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx],
                                                  Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx])
    
    if 1:
      azNext,elNext = azel_fromRotMatrix_dir(Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Direction_Next'][itx],dir)
      gainNext = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternMaxGain'][itx]
      gainNext*=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(azNext), np.rad2deg(elNext), 
                                                  1, Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternAzimuthBeamWidth'][itx],
                                                    Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternElevationBeamWidth'][itx],
                                                    Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-PatternType'][itx])

  Total_Transmit_Power = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Power'][itx]
  PTX = gain * Total_Transmit_Power
  amp*=np.sqrt(PTX)
  amp/=distanceLP*np.sqrt(4*np.pi)
  # Amplitude 2
  if next_element.source_type=='scatter':
    area = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
    RCS0 = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
    area*=np.abs(next_normal.dot(-dir))
    rcs = np.sqrt(area * RCS0)
    amp*=np.sqrt(rcs)
  # print(amp,"1,2",rcs,"RCS")
  if next_element.source_type=='RIS':
    isuiteRIS = next_element.ids[0]
    iris = next_element.ids[1]
    iriselement = next_element.ids[2]
    if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
      gain = 1
    else:
      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
      gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
      gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

    Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
    Aeff = gain * Lambda * Lambda / (4 * np.pi)
    amp*=np.sqrt(Aeff)
  
  return [d,drate,amp,Scatters_in_the_Middle],False

def calc_Path_Middle(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle):
  for imiddle_elements in range(len(path.middle_elements)-1):
    middle_element = path.middle_elements[len(path.middle_elements)-1-imiddle_elements]
    if middle_element.source_type=='scatter':
      p = middle_element.ids[1]
      currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
      currentPoint = Vector3D(p.x,p.y,p.z)
      normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
      area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
      RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
      next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
      if next_element.source_type=='scatter':
        next = next_element.ids[1]
        nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
        next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
        Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])

        Next = Vector3D(next.x,next.y,next.z)
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        area*=np.abs(normal.dot(dir))
        rcs = np.sqrt(area * RCS0)
        amp*=np.sqrt(rcs)
        amp/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        area     = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
        area*=np.abs(next_normal.dot(-dir))
        RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
        rcs = np.sqrt(area * RCS0)
        amp*=np.sqrt(rcs)

      if next_element.source_type=='RIS':
        isuiteRIS = next_element.ids[0]
        iris = next_element.ids[1]
        iriselement = next_element.ids[2]
        next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
        Next = Vector3D(next.x,next.y,next.z)
        nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        area*=np.abs(normal.dot(dir))
        rcs = np.sqrt(area * RCS0)
        amp*=np.sqrt(rcs)
        amp/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
          gain = 1
        else:
          az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
          gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
          gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

        Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        Aeff = gain * Lambda * Lambda / (4 * np.pi)
        amp*=np.sqrt(Aeff)
        
            
        risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
        risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        risd *= (math.fmod(risphase,360)/360)
        amp*=np.sqrt(risamp)
        d+=risd
    if middle_element.source_type=='RIS':
      isuiteRIS0 = middle_element.ids[0]
      iris0 = middle_element.ids[1]
      iriselement0 = middle_element.ids[2]
      p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
      currentPoint = Vector3D(p.x,p.y,p.z)
      currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
      next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
      if next_element.source_type=='scatter':
        next = next_element.ids[1]
        nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
        next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
        Next = Vector3D(next.x,next.y,next.z)
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
          gain = 1
        else:
          az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
          gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
          gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
        amp*=np.sqrt(gain)
        amp/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        area     = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][3]
        area*=np.abs(next_normal.dot(-dir))
        RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][5]
        rcs = np.sqrt(area * RCS0)
        amp*=np.sqrt(rcs)

      if next_element.source_type=='RIS':
        isuiteRIS = next_element.ids[0]
        iris = next_element.ids[1]
        iriselement = next_element.ids[2]
        next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
        Next = Vector3D(next.x,next.y,next.z)
        nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
          gain = 1
        else:
          az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
          gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
          gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
        amp*=np.sqrt(gain)
        amp/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
          gain = 1
        else:
          az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
          gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
          gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

        Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        Aeff = gain * Lambda * Lambda / (4 * np.pi)
        amp*=np.sqrt(Aeff)
            
        risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
        risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        risd *= (math.fmod(risphase,360)/360)
        amp*=np.sqrt(risamp)
        d+=risd
  return d,drate,amp,Scatters_in_the_Middle



def calc_Path_RX(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp):
    middle_element = path.middle_elements[0]
    isuiteRX = path.receiver[0]
    iradarRX = path.receiver[1]
    irx = path.receiver[2]
    next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
    Next = Vector3D(next.x,next.y,next.z)
    nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]
    if middle_element.source_type=='scatter':
      p = middle_element.ids[1]
      currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
      currentPoint = Vector3D(p.x,p.y,p.z)
      normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
      area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
      RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
    if middle_element.source_type=='RIS':
      isuiteRIS0 = middle_element.ids[0]
      iris0 = middle_element.ids[1]
      iriselement0 = middle_element.ids[2]
      p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
      currentPoint = Vector3D(p.x,p.y,p.z)
      currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
        
      risamp,risphase=Suite_Position[isuiteRIS0]['RIS'][iris0]['PhaseAmplitude'][iriselement0]
      risd=Suite_Position[isuiteRIS0]['RIS'][iris0]['WaveLength']
      risd *= (math.fmod(risphase,360)/360)
      amp*=np.sqrt(risamp)
      d+=risd
      # next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
    dir = next-p
    distanceLP =  dir.magnitude
    dir/= distanceLP
    # Distance
    d+=Next.distance_to(currentPoint)
    # DistanceRate
    drate += (-currentv.dot(dir)+nextv.dot(dir))
    # Amplitude 1
    if middle_element.source_type=='scatter':
      area*=np.abs(normal.dot(dir))
      rcs = np.sqrt(area * RCS0)
      amp*=np.sqrt(rcs)
    # print(amp,"1,3",rcs,"RCS")
    if middle_element.source_type=='RIS':
      if Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternType'][iriselement0]=='Omni':
        gain = 1
      else:
        az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS0]['RIS'][iris0]['Direction'][iriselement0],dir)
        gain = Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternMaxGain'][iriselement0]
        gain*= np.sinc(az/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternAzimuthBeamWidth'][iriselement0])*np.sinc(el/Suite_Position[isuiteRIS0]['RIS'][iris0]['PatternElevationBeamWidth'][iriselement0])
      amp*=np.sqrt(gain)

    amp/=distanceLP*np.sqrt(4*np.pi)
    # Amplitude 2
    if Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx]=='Omni':
      gain = 1
    else:
      az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Direction'][irx],-dir)
      gain = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternMaxGain'][irx]
      gain = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternMaxGain'][irx]
      gain*=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(az), np.rad2deg(el), 
                                                1, Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx],
                                                  Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx],
                                                  Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternType'][irx])
      
      # gain*= np.sinc(az/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternAzimuthBeamWidth'][irx])*np.sinc(el/Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-PatternElevationBeamWidth'][irx])
    Lambda = Suite_Position[isuiteRX]['Radar'][iradarRX]['WaveLength']
    Aeff = gain * Lambda * Lambda / (4 * np.pi)
    amp*=np.sqrt(Aeff)
    return d,drate,amp


def paths_processing(Paths,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,all_d_drate_amp):
  for path in Paths:
    # if len(path.middle_elements)==3:
    #   1
    amp = np.longdouble(1)
    d = np.longdouble(0)
    drate = 0
    # DESIRED FOR EACH PATH: AMP D DRATE

    # if len(path.middle_elements)==0: # Strong Interference + Friis Equation
    #   continue
    d_drate_amp_Scatters_in_the_Middle,is_end=calc_Path_TX_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp)
    if is_end:
      RXfromTX = ssp.config.directReceivefromTX
      if RXfromTX:
        if amp==0:
          continue
        all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append(d_drate_amp_Scatters_in_the_Middle)
      continue
    d,drate,amp,Scatters_in_the_Middle = d_drate_amp_Scatters_in_the_Middle
    if amp==0:
      continue
    
    # Middle Elements
    # d,drate,amp,Scatters_in_the_Middle=calc_Path_Middle(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle)
    d,drate,amp,Scatters_in_the_Middle=calc_Path_Middle_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle)
    if amp==0:
      continue
    
    # d,drate,amp=calc_Path_RX(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp)
    d,drate,amp=calc_Path_RX_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp)
    if amp==0:
      continue
    all_d_drate_amp[path.receiver[0]][path.receiver[1]][path.receiver[2]][path.transmitter[0]][path.transmitter[1]][path.transmitter[2]].append([d,drate,amp,Scatters_in_the_Middle])

def tx2rx(path,Suite_Position,d,drate,amp):
  
  isuiteTX = path.transmitter[0]
  iradarTX = path.transmitter[1]
  itx = path.transmitter[2]
  tx = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Position'][itx]
  tx_v = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Velocity'][itx]
  TX = Vector3D(tx.x,tx.y,tx.z)
  
  isuiteRX = path.receiver[0]
  iradarRX = path.receiver[1]
  irx = path.receiver[2]
  next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
  Next = Vector3D(next.x,next.y,next.z)
  nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]

  dir = next-tx
  distanceLP =  dir.magnitude
  dir/= distanceLP

  # Distance
  d+=Next.distance_to(TX)
  # DistanceRate
  drate += (-tx_v.dot(dir)+nextv.dot(dir))
  
  amp=tx_amp(distanceLP,Suite_Position[isuiteTX]['Radar'][iradarTX],itx,amp,dir)
  amp=rx_amp(distanceLP,Suite_Position[isuiteRX]['Radar'][iradarRX],irx,amp,-dir)
  return d,drate,amp

def tx2scatter_or_ris(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle):
  isuiteTX = path.transmitter[0]
  iradarTX = path.transmitter[1]
  itx = path.transmitter[2]
  tx = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Position'][itx]
  tx_v = Suite_Position[isuiteTX]['Radar'][iradarTX]['TX-Velocity'][itx]
  TX = Vector3D(tx.x,tx.y,tx.z)
  
  # find Next for TX (Step 0)
  next_element = path.middle_elements[len(path.middle_elements)-1]
  if next_element.source_type=='scatter':
    next = next_element.ids[1]
    nextv = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
    # next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
    Next = Vector3D(next.x,next.y,next.z)
    Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])
  if next_element.source_type=='RIS':
    isuiteRIS = next_element.ids[0]
    iris = next_element.ids[1]
    iriselement = next_element.ids[2]
    next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
    Next = Vector3D(next.x,next.y,next.z)
    nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
    
    risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
    risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
    risd *= (math.fmod(risphase,360)/360)/2
    amp*=np.sqrt(np.sqrt(risamp))
    d+=risd
    # print(next_element.ids,next,nextv)
    Scatters_in_the_Middle.append(f'RIS_{isuiteRIS}_{iris}_{iriselement}')
  
  dir = next-tx
  distanceLP =  dir.magnitude
  dir/= distanceLP
  # Distance
  d+=Next.distance_to(TX)
  # DistanceRate
  drate += (-tx_v.dot(dir)+nextv.dot(dir))
  # Amplitude 1
  amp=tx_amp(distanceLP,Suite_Position[isuiteTX]['Radar'][iradarTX],itx,amp,dir)
  
  # Amplitude 2
  if next_element.source_type=='scatter':
    amp=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]],-dir,amp)
  if next_element.source_type=='RIS':
    isuiteRIS = next_element.ids[0]
    iris = next_element.ids[1]
    iriselement = next_element.ids[2]
    amp=rx_amp(distanceLP,Suite_Position[isuiteRIS]['RIS'][iris],iriselement,amp,-dir)
  return d,drate,amp
    # if Suite_Position[isuiteRIS]['RIS'][iris]['PatternType'][iriselement]=='Omni':
    #   gain = 1
    # else:
    #   az,el = azel_fromRotMatrix_dir(Suite_Position[isuiteRIS]['RIS'][iris]['Direction'][iriselement],dir)
    #   gain = Suite_Position[isuiteRIS]['RIS'][iris]['PatternMaxGain'][iriselement]
    #   gain*= np.sinc(az/Suite_Position[isuiteRIS]['RIS'][iris]['PatternAzimuthBeamWidth'][iriselement])*np.sinc(el/Suite_Position[isuiteRIS]['RIS'][iris]['PatternElevationBeamWidth'][iriselement])

    # Lambda = Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
    # Aeff = gain * Lambda * Lambda / (4 * np.pi)
    # amp*=np.sqrt(Aeff)
  #   def rx_amp(radar,irx,amp):
  # gain = antenna_gain(radar['RX-PatternType'][irx],
  #                     radar['RX-Direction'][irx],
  #                     radar['RX-PatternMaxGain'][irx], 
  #                     radar['RX-PatternAzimuthBeamWidth'][irx],
  #                     radar['RX-PatternElevationBeamWidth'][irx],
  #                     radar['RX-PatternType'][irx])
  # Lambda = radar['WaveLength']
  # Aeff = gain * Lambda * Lambda / (4 * np.pi)
  # amp*=np.sqrt(Aeff)


def scatter_amp(distanceLP,SG,dir,amp):
  area = SG[3]
  normal = SG[4]
  RCS0 = SG[5]
  area*=np.abs(normal.dot(-dir))
  rcs = np.sqrt(area * RCS0)
  co = 1.0/ ( np.sqrt(4*np.pi) * distanceLP )
  amp*=np.sqrt(co*rcs)
  return amp
def tx_amp(distanceLP,radar,itx,amp,dir):
  gain = antenna_gain(radar['TX-PatternType'][itx],
                      radar['TX-Direction'][itx],
                      radar['TX-PatternMaxGain'][itx], 
                      radar['TX-PatternAzimuthBeamWidth'][itx],
                      radar['TX-PatternElevationBeamWidth'][itx],
                      radar['TX-PatternType'][itx],dir)
  Total_Transmit_Power = radar['TX-Power'][itx]
  PTX = gain * Total_Transmit_Power
  co = 1.0/ ( np.sqrt(4*np.pi) * distanceLP )
  amp*=np.sqrt( co * PTX)
  return amp
def antenna_gain(PatternType,antenna_direction,PatternMaxGain, azimuth_beamwidth, elevation_beamwidth,antennaType,dir):
  if PatternType=='Omni':
    gain = 1
  else:
    az,el = azel_fromRotMatrix_dir(antenna_direction,dir)
    gain=ssp.radar.utils.antenna_gain_from_beamwidth(np.rad2deg(az), np.rad2deg(el),PatternMaxGain, azimuth_beamwidth, elevation_beamwidth,antennaType)
    # gain = np.rad2deg(az)
  return gain
def rx_amp(distanceLP,radar,irx,amp,dir):
  if 'RX-PatternType' in radar:
    gain = antenna_gain(radar['RX-PatternType'][irx],
                      radar['RX-Direction'][irx],
                      radar['RX-PatternMaxGain'][irx], 
                      radar['RX-PatternAzimuthBeamWidth'][irx],
                      radar['RX-PatternElevationBeamWidth'][irx],
                      radar['RX-PatternType'][irx],dir)
  else:
    gain = antenna_gain(radar['PatternType'][irx],
                      radar['Direction'][irx],
                      radar['PatternMaxGain'][irx], 
                      radar['PatternAzimuthBeamWidth'][irx],
                      radar['PatternElevationBeamWidth'][irx],
                      radar['PatternType'][irx],dir)
  Lambda = radar['WaveLength']
  Aeff = gain * Lambda * Lambda / (4 * np.pi)
  co = 1.0/ ( np.sqrt(4*np.pi) * distanceLP )
  amp*=np.sqrt(co *Aeff)
  return amp
def tx_amp_ris(distanceLP,radar,irx,amp,dir):
  gain = antenna_gain(radar['PatternType'][irx],
                      radar['Direction'][irx],
                      radar['PatternMaxGain'][irx], 
                      radar['PatternAzimuthBeamWidth'][irx],
                      radar['PatternElevationBeamWidth'][irx],
                      radar['PatternType'][irx],dir)  
  # co = 1.0/ ( np.sqrt(4*np.pi) * distanceLP )
  amp*=np.sqrt(gain/ np.sqrt(( (4 * np.pi) * distanceLP**2 )))
  # G_ris / np.sqrt(( (4 * np.pi) * R_ris_scatter**2 ))
  return amp

def calc_Path_TX_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp):
  Scatters_in_the_Middle = []
  if len(path.middle_elements)==0:
    d,drate,amp=tx2rx(path,Suite_Position,d,drate,amp)
    return [d,drate,amp,Scatters_in_the_Middle],True
  d,drate,amp=tx2scatter_or_ris(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle)
  return [d,drate,amp,Scatters_in_the_Middle],False

def calc_Path_Middle_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp,Scatters_in_the_Middle):
  for imiddle_elements in range(len(path.middle_elements)-1):
    middle_element = path.middle_elements[len(path.middle_elements)-1-imiddle_elements]
    if middle_element.source_type=='scatter':
      p = middle_element.ids[1]
      currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
      currentPoint = Vector3D(p.x,p.y,p.z)
      # normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
      # area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
      # RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
      next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
      if next_element.source_type=='scatter':
        next = next_element.ids[1]
        nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
        # next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
        Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])
        Next = Vector3D(next.x,next.y,next.z)
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        amp0=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]],dir,1.0)    
        # amp0/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        amp0=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]],-dir,amp0)    
        amp *= amp0
      if next_element.source_type=='RIS':
        isuiteRIS = next_element.ids[0]
        iris = next_element.ids[1]
        iriselement = next_element.ids[2]
        Scatters_in_the_Middle.append(f'RIS_{isuiteRIS}_{iris}_{iriselement}')
        next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
        Next = Vector3D(next.x,next.y,next.z)
        nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        amp0=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]],dir,1.0)    
        # amp0/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        amp0=rx_amp(distanceLP,Suite_Position[isuiteRIS]['RIS'][iris],iriselement,amp0,-dir)
        amp *= amp0
                #RIS
        risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
        risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        risd *= (math.fmod(risphase,360)/360)/2
        amp*=np.sqrt(np.sqrt(risamp))
        d+=risd
    if middle_element.source_type=='RIS':
      isuiteRIS0 = middle_element.ids[0]
      iris0 = middle_element.ids[1]
      iriselement0 = middle_element.ids[2]
      p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
      currentPoint = Vector3D(p.x,p.y,p.z)
      currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
      next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
      if next_element.source_type=='scatter':
        next = next_element.ids[1]
        nextv            = ScattersGeoV[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]]
        # next_normal = ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]][4]
        Next = Vector3D(next.x,next.y,next.z)
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        amp0=tx_amp_ris(distanceLP,Suite_Position[isuiteRIS0]['RIS'][iris0],iriselement0,1.0,dir)
        # Amplitude 2
        
        amp0=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]]],-dir,amp0)
        amp *= amp0
        Scatters_in_the_Middle.append([HashFaceIndex_ScattersGeo[next_element.ids[0][1]][next_element.ids[0][2]],next])
        

      if next_element.source_type=='RIS':
        isuiteRIS = next_element.ids[0]
        iris = next_element.ids[1]
        iriselement = next_element.ids[2]
        Scatters_in_the_Middle.append(f'RIS_{isuiteRIS}_{iris}_{iriselement}')
        
        next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
        Next = Vector3D(next.x,next.y,next.z)
        nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
        dir = next-p
        distanceLP =  dir.magnitude
        dir/= distanceLP
        # Distance
        d+=Next.distance_to(currentPoint)
        # DistanceRate
        drate += (-currentv.dot(dir)+nextv.dot(dir))
        # Amplitude 1
        amp0=tx_amp_ris(distanceLP,Suite_Position[isuiteRIS0]['RIS'][iris0],iriselement0,1.0,dir)
        # amp0/=distanceLP*np.sqrt(4*np.pi)
        # Amplitude 2
        amp0=rx_amp(distanceLP,Suite_Position[isuiteRIS]['RIS'][iris],iriselement,amp0,-dir)
        amp *= amp0
        #RIS    
        risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
        risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
        risd *= (math.fmod(risphase,360)/360)/2
        amp*=np.sqrt(np.sqrt(risamp))
        d+=risd

      #RIS
      risamp,risphase=Suite_Position[isuiteRIS0]['RIS'][iris0]['PhaseAmplitude'][iriselement0]
      risd=Suite_Position[isuiteRIS0]['RIS'][iris0]['WaveLength']
      risd *= (math.fmod(risphase,360)/360)/2
      amp*=np.sqrt(np.sqrt(risamp))
      d+=risd
  
  return d,drate,amp,Scatters_in_the_Middle
  
def calc_Path_RX_functional(path,Suite_Position,ScattersGeoV,HashFaceIndex_ScattersGeo,ScattersGeo,d,drate,amp):
    middle_element = path.middle_elements[0]
    isuiteRX = path.receiver[0]
    iradarRX = path.receiver[1]
    irx = path.receiver[2]
    next = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Position'][irx]
    Next = Vector3D(next.x,next.y,next.z)
    nextv = Suite_Position[isuiteRX]['Radar'][iradarRX]['RX-Velocity'][irx]
    if middle_element.source_type=='scatter':
      p = middle_element.ids[1]
      currentv = ScattersGeoV[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]]
      currentPoint = Vector3D(p.x,p.y,p.z)
      normal = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][4]
      area     = ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][3]
      RCS0 =     ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]][5]
    if middle_element.source_type=='RIS':
      isuiteRIS0 = middle_element.ids[0]
      iris0 = middle_element.ids[1]
      iriselement0 = middle_element.ids[2]
      p = Suite_Position[isuiteRIS0]['RIS'][iris0]['Position'][iriselement0]
      currentPoint = Vector3D(p.x,p.y,p.z)
      currentv = Suite_Position[isuiteRIS0]['RIS'][iris0]['Velocity'][iriselement0]
      risamp,risphase=Suite_Position[isuiteRIS0]['RIS'][iris0]['PhaseAmplitude'][iriselement0]
      risd=Suite_Position[isuiteRIS0]['RIS'][iris0]['WaveLength']
      risd *= (math.fmod(risphase,360)/360)/2
      amp*=np.sqrt(np.sqrt(risamp))
      d+=risd
      # next_element   = path.middle_elements[len(path.middle_elements)-1-imiddle_elements-1]
    dir = next-p
    distanceLP =  dir.magnitude
    dir/= distanceLP
    # Distance
    d+=Next.distance_to(currentPoint)
    # DistanceRate
    drate += (-currentv.dot(dir)+nextv.dot(dir))
    # Amplitude 1
    if middle_element.source_type=='scatter':
      amp=scatter_amp(distanceLP,ScattersGeo[HashFaceIndex_ScattersGeo[middle_element.ids[0][1]][middle_element.ids[0][2]]],dir,amp)
    if middle_element.source_type=='RIS':
      amp=tx_amp_ris(distanceLP,Suite_Position[isuiteRIS0]['RIS'][iris0],iriselement0,amp,dir)
        
      # amp=tx(distanceLP,Suite_Position[isuiteRIS0]['RIS'][iris0],iriselement0,amp)
    # amp/=distanceLP*np.sqrt(4*np.pi)
    
    # Amplitude 2
    amp=rx_amp(distanceLP,Suite_Position[isuiteRX]['Radar'][iradarRX],irx,amp,-dir)
  
    return d,drate,amp

# def path_element_processing(next_element,SG,H,Suite_Position):
#   if next_element.source_type=='scatter':
#     next = next_element.ids[1]
#     nextv = SG[H[next_element.ids[0][1]][next_element.ids[0][2]]]
#     next_normal = SG[H[next_element.ids[0][1]][next_element.ids[0][2]]][4]
#     Next = Vector3D(next.x,next.y,next.z)
#     Scatters_in_the_Middle.append(H[next_element.ids[0][1]][next_element.ids[0][2]])
#   if next_element.source_type=='RIS':
#     isuiteRIS = next_element.ids[0]
#     iris = next_element.ids[1]
#     iriselement = next_element.ids[2]
#     next = Suite_Position[isuiteRIS]['RIS'][iris]['Position'][iriselement]
#     Next = Vector3D(next.x,next.y,next.z)
#     nextv = Suite_Position[isuiteRIS]['RIS'][iris]['Velocity'][iriselement]
    
#     risamp,risphase=Suite_Position[isuiteRIS]['RIS'][iris]['PhaseAmplitude'][iriselement]
#     risd=Suite_Position[isuiteRIS]['RIS'][iris]['WaveLength']
#     risd *= (math.fmod(risphase,360)/360)/2
#     amp*=np.sqrt(np.sqrt(risamp))
#     d+=risd
#     # print(next_element.ids,next,nextv)
  


import sensingsp as ssp
import matplotlib.pyplot as plt
import numpy as np
import bpy
import os
import tempfile
def define_settings():
    if not("Simulation Settings" in bpy.data.objects):
        bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=(0, 0, 0), rotation=(0, 0, 0), scale=(.01, .01, .01))
        sim_axes = bpy.context.object
        sim_axes.name = f'Simulation Settings'
        sim_axes["RF Frequency (GHz)"] = 70
        sim_axes["Velocity Calc. Jump"] = 1
        # sim_axes["Velocity Calc. Jump"] = bpy.props.IntProperty(name="Velocity Calc. Jump", default=1, min=1)
        sim_axes["Bounce Number"] = 2
        sim_axes["Render Blender Frames"] = True#False
        sim_axes["Open Output Folder"] = True
        sim_axes["do RayTracing LOS"] = True
        sim_axes["do RayTracing Simple"] = False
        sim_axes["CUDA SignalGeneration Enabled"] = False
        sim_axes["Debug_BypassCPITiming"] = False
        
        
        # bpy.data.objects["Simulation Settings"]["do RayTracing LOS"]=False
       
        
        # bpy.types.Object.velocity_calc_jump = bpy.props.IntProperty(name="Velocity Calc. Jump", default=1, min=1)
        # bpy.types.Object.bounce_number = bpy.props.IntProperty(name="Bounce Number", default=2, min=2)
        # bpy.types.Object.render_blender_frames = bpy.props.BoolProperty(name="Render Blender Frames", default=True)
        # bpy.types.Object.open_output_folder = bpy.props.BoolProperty(name="Open Output Folder", default=True)

        
        temp_dir = tempfile.gettempdir()
        radarsim_dir = os.path.join(temp_dir, 'SensingSP')
        os.makedirs(radarsim_dir, exist_ok=True)
        sim_axes["Video Directory"]=radarsim_dir
        sim_axes["Add Ris"] = f"0,8,8,{os.path.join(radarsim_dir, 'Ris.mat')},2"
        sim_axes["Add Probe"] = f"1,.1,20,20,db1,range_effect1,add_color1,add_open3d1,colormap1,0.05"


def blender_buttons_run_2():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.define_axes(2)
    ssp.config.setDopplerProcessingMethod_FFT_Winv(1)
    while ssp.config.run():
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
        ssp.integratedSensorSuite.SensorsSignalProccessing_Angle_frame(Signals)
        ssp.utils.increaseCurrentFrame()


import bpy
import numpy as np
from mathutils import Vector
from scipy.io import loadmat

def add_ris(isuite=0, iris=0, location=Vector((0, 0, 0)), rotation=Vector((np.pi/2, 0, -np.pi/2)), f0=70e9, N1=8, N2=8,
              passive_active='passive',steeringPhase_az_el=[0,0]):
    fromFile = False
    if "Simulation Settings" in bpy.data.objects:
        option = bpy.data.objects["Simulation Settings"]["Add Ris"]
        parts = option.split(',')
        # Convert parts to appropriate types
        input_or_File = int(parts[0])
        if input_or_File == 1:
            N1 = int(parts[1])
            N2 = int(parts[2])
        if input_or_File == 2:
            file = parts[3]
            data = loadmat(file)
            coordinates = data["P"]
            amplitude  = data["amp"]
            phase = data["phase"]
            fromFile = True
    Lambda = 3e8 / f0
    Suitename = f'SuitePlane_{isuite}'
    Suite_obj = bpy.data.objects[Suitename]

    bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=location, rotation=rotation, scale=(1, 1, 1))
    empty = bpy.context.object
    empty.name = f'RISPlane_{isuite}_{iris}'
    empty["Center_Frequency_GHz"] = 70
    empty["passive active"] = passive_active
    empty["steeringPhase_az_el"] = steeringPhase_az_el
    # empty["moein_boolean"] = True
    empty.parent = Suite_obj

    s = 0.05
    Type = 'AREA'
    
    if fromFile:
        for i in range(coordinates.shape[0]):
            bpy.ops.object.light_add(type=Type, radius=1, location=(coordinates[i,0],coordinates[i,1],coordinates[i,2]))
            tx = bpy.context.object
            tx.scale = (s*Lambda/2, s*Lambda/2, s*Lambda/2)
            NF = 80
            tx.name = f'RIS_Element_{isuite}_{iris}_{i}_{0}_{NF}'
            tx['amplitude']=amplitude[0,i]/1.0
            tx['phase']=phase[0,i]/1.0
            tx.parent = empty

    else:
        for i1 in range(N1):
            for i2 in range(N2):
                bpy.ops.object.light_add(type=Type, radius=1, location=(i1*Lambda/2, i2*Lambda/2, 0))
                tx = bpy.context.object
                tx.scale = (s*Lambda/2, s*Lambda/2, s*Lambda/2)
                NF = 80
                tx.name = f'RIS_Element_{isuite}_{iris}_{i1}_{i2}_{NF}'
                tx['amplitude']=1
                tx['phase']=0
                tx.parent = empty

def add_ris_mat(isuite=0, iris=0, location=Vector((0, 0, 0)), rotation=Vector((np.pi/2, 0, -np.pi/2)), f0=70e9, pos=np.array([]),passive_active='passive'):
    Lambda = 3e8 / f0
    Suitename = f'SuitePlane_{isuite}'
    Suite_obj = bpy.data.objects[Suitename]

    bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=location, rotation=rotation, scale=(1, 1, 1))
    empty = bpy.context.object
    empty.name = f'RISPlane_{isuite}_{iris}'
    empty["Center_Frequency_GHz"] = 70
    empty["passive active"] = passive_active
    empty["steeringPhase_az_el"] = []
    empty.parent = Suite_obj

    s = 0.05
    Type = 'AREA'
    N1=pos.shape[0]
    for i1 in range(N1):
        bpy.ops.object.light_add(type=Type, radius=1, location=(pos[i1,0], pos[i1,1], 0))
        # bpy.ops.object.empty_add(type='PLAIN_AXES', location=(pos[i1, 0], pos[i1, 1], 0))
        tx = bpy.context.object
        tx.scale = (s*Lambda/2, s*Lambda/2, s*Lambda/2)
        NF = 80
        tx.name = f'RIS_Element_{isuite}_{iris}_{i1}_{0}_{NF}'
        tx['amplitude']=pos[i1,4]
        tx['phase']=pos[i1,3]
        tx.parent = empty




import numpy as np
from scipy.linalg import hadamard
import scipy.special as sp
import scipy.stats as stats
import scipy.signal.windows as scipy_windows
import math
import numba

class MIMO_Functions:
  
    @numba.jit(nopython=True)
    def sv(self, az, fd, W):
        NPulse = W.shape[0]
        M = W.shape[1]
        a = np.exp(1j * np.pi * np.arange(M) * np.sin(np.deg2rad(az)))
        d = np.exp(1j * 2 * np.pi * np.arange(NPulse) * fd)
        s = (W @ a) * d
        return s

    @numba.jit(nopython=True)
    def AD_matrix(self, NPulse, M, tech='TDM'):
        W = np.zeros((NPulse, M), dtype=np.complex128)
        if tech == "DDM":
            for p in range(NPulse):
                for m in range(M):
                    W[p, m] = np.exp(1j * 2 * np.pi * m * p / M)
        elif tech == "DDM2":
            for p in range(NPulse):
                for m in range(M):
                    W[p, m] = np.exp(1j * 2 * np.pi * m * p / NPulse)
        elif tech == "DDM_RandPhase":
            for p in range(NPulse):
                for m in range(M):
                    W[p, m] = np.exp(1j * 2 * np.pi * np.random.rand())
        elif tech == "TDM":
            for p in range(NPulse):
                for m in range(M):
                    W[p, m] = 0
                    if p % M == m:
                        W[p, m] = 1
        elif tech == "BPM":
            H = hadamard(M, dtype=complex)
            for p in range(NPulse):
                for m in range(M):
                    W[p, m] = H[p % M, m]
        else:
            raise ValueError("Not Valid!")
        return W

@numba.jit(nopython=True)
def generate_complex_gaussian_samples(cholesky_covariance_matrix, K):
    N = cholesky_covariance_matrix.shape[0]
    real_part = np.random.randn(N, K)
    imag_part = np.random.randn(N, K)
    Z = (real_part + 1j * imag_part) / np.sqrt(2)
    samples = cholesky_covariance_matrix @ Z
    return samples

@numba.jit(nopython=True)
def array_detector_ABC(A, B, C):
    return A @ B @ C

@numba.jit(nopython=True)
def array_detector_ABC_inv(A, B, C):
    return np.linalg.pinv(array_detector_ABC(A, B, C))

@numba.jit(nopython=True)
def array_detector_Hermitian(A):
    return np.conj(A.T)

@numba.jit(nopython=True)
def array_detector_MF(x, Rinv, S):
    v = array_detector_ABC(array_detector_Hermitian(S), Rinv, x)
    A = array_detector_ABC(array_detector_Hermitian(S), Rinv, S)
    A = np.linalg.pinv(A)
    t = array_detector_ABC(array_detector_Hermitian(v), A, v)[0, 0]
    return np.abs(t)

@numba.jit(nopython=True)
def MF_threshold(Q, pfa):
    # Q is the degrees of freedom (related to the number of samples), pfa is the desired false alarm probability
    return stats.chi2.ppf(1 - pfa, 2 * Q) / 2

@numba.jit(nopython=True)
def MF_pfa(Q, threshold):
    return 1 - stats.chi2.cdf(threshold * 2, 2 * Q)

@numba.jit(nopython=True)
def MF_pd(Q, threshold, SCINR_Post):
    return 1 - stats.ncx2.cdf(x=2 * threshold, df=2 * Q, nc=2 * SCINR_Post)

@numba.jit(nopython=True)
def array_detector_ED(x, Rinv):
    t = array_detector_ABC(array_detector_Hermitian(x), Rinv, x)[0, 0]
    return np.abs(t)

@numba.jit(nopython=True)
def array_detector_ACE(x, Rinv, S):
    return array_detector_MF(x, Rinv, S) / array_detector_ED(x, Rinv)

@numba.jit(nopython=True)
def array_detector_Kelly(x, Rinv, S):
    return array_detector_MF(x, Rinv, S) / (1 + array_detector_ED(x, Rinv))

@numba.jit(nopython=True)
def Kelly_pfa(Q, NSD, NDim, threshold):
    return 1 - sp.betainc(Q, NSD - NDim + 1, threshold)

@numba.jit(nopython=True)
def Kelly_threshold(Q, NSD, NDim, pfa):
    return sp.betaincinv(Q, NSD - NDim + 1, 1 - pfa)

@numba.jit(nopython=True)
def array_detector_Rao(x, R, S):
    Rinv = np.linalg.pinv(R + array_detector_Hermitian(x) @ x)
    v = array_detector_ABC(array_detector_Hermitian(S), Rinv, x)
    A = array_detector_ABC(array_detector_Hermitian(S), Rinv, S)
    A = np.linalg.pinv(A)
    t = array_detector_ABC(array_detector_Hermitian(v), A, v)[0, 0]
    return np.abs(t)

@numba.jit(nopython=True)
def array_detector_MMED(R):
    eigenvalues = np.linalg.eigvals(R)
    return np.abs(np.max(eigenvalues) / np.min(eigenvalues))

@numba.jit(nopython=True)
def array_detector_MED(x, R):
    eigenvalues = np.linalg.eigvals(R)
    e = array_detector_Hermitian(x) @ x
    return np.abs(e[0, 0] / np.min(eigenvalues))

@numba.jit(nopython=True)
def array_detector_RankOne_SingleShot(X):
    R = X @ array_detector_Hermitian(X)
    eigenvalues = np.linalg.eigvals(R)
    return np.abs(np.max(eigenvalues) / np.trace(R))

@numba.jit(nopython=True)
def array_detector_SCM_N(X):
    return X @ array_detector_Hermitian(X)

@numba.jit(nopython=True)
def array_detector_cos2(S, Sp, Rinv):
    A = array_detector_ABC(array_detector_Hermitian(S), Rinv, S)
    B = array_detector_ABC(array_detector_Hermitian(S), Rinv, Sp)
    BhAinvB = array_detector_ABC(array_detector_Hermitian(B), np.linalg.pinv(A), B)
    C = array_detector_ABC(array_detector_Hermitian(Sp), Rinv, Sp)
    ones = np.ones((S.shape[1], 1))
    num = array_detector_Hermitian(ones) @ BhAinvB @ ones
    denum = array_detector_Hermitian(ones) @ C @ ones
    return np.abs(num[0, 0] / denum[0, 0])

@numba.jit(nopython=True)
def array_detector_SINR(S, alpha, Rinv):
    A = array_detector_ABC(array_detector_Hermitian(S), Rinv, S)
    SINR = array_detector_ABC(array_detector_Hermitian(alpha), A, alpha)
    return np.abs(SINR[0, 0])

@numba.jit(nopython=True)
def array_detector_ROC_MonteCarlo_sortxH0(xH0, xH1, N=200):
    THR_v = xH0   
    THR_v = THR_v[THR_v.argsort()]
    pfa_v = np.logspace(0, -np.log10(THR_v.shape[0]), N)
    indices = []
    Pd_o  = []
    Pfa_o = []
    for pfai in pfa_v:
        ind = round((1 - pfai) * (THR_v.shape[0] - 1))
        if ind not in indices:
            indices.append(ind)
            THR = THR_v[ind]
            pfa = np.sum(xH0 > THR) / THR_v.shape[0]
            pd = np.sum(xH1 > THR) / THR_v.shape[0]
            Pd_o.append(pd)
            Pfa_o.append(pfa)
    return np.array(Pfa_o), np.array(Pd_o)

@numba.jit(nopython=True)
def array_detector_ROC_AMF_ACE_MonteCarlo(Min_Pfa, S, SigmaSQRT, NSD, alpha0_SINR, NPfa=100, progbar=True):
    t_mc = []
    alpha = alpha0_SINR * np.ones((S.shape[1], 1))
    signal_H1_0 = S @ alpha
    N_MC_1 = int(NPfa / Min_Pfa)
    mc_100 = 0
    for mc in range(N_MC_1):
        if progbar:
            if mc_100 < mc * 100 / N_MC_1:
                mc_100 += 1
                print(mc_100)
        noiseSD = generate_complex_gaussian_samples(SigmaSQRT, NSD)
        NSigmaHat = noiseSD @ np.conj(noiseSD.T)
        NDim = NSigmaHat.shape[0]
        NSigmaHat_DL = NSigmaHat + 0 * np.eye(NDim)
        NSigmaHat_DL_inv = np.linalg.pinv(NSigmaHat_DL)
        N_Trick_MC = 1
        for _ in range(N_Trick_MC):
            noiseH0 = generate_complex_gaussian_samples(SigmaSQRT, 1)
            signal_H0 = noiseH0
            t0_ACE = array_detector_ACE(signal_H0, NSigmaHat_DL_inv, S)
            t0_AMF = array_detector_MF(signal_H0, NSigmaHat_DL_inv, S)
            noiseH1 = generate_complex_gaussian_samples(SigmaSQRT, 1)
            signal_H1 = noiseH1 + signal_H1_0 
            t1_ACE = array_detector_ACE(signal_H1, NSigmaHat_DL_inv, S)
            t1_AMF = array_detector_MF(signal_H1, NSigmaHat_DL_inv, S)
            t_mc.append([t0_AMF, t0_ACE, t1_AMF, t1_ACE])
    t_mc = np.array(t_mc)
    Pfa_AMF, Pd_AMF = array_detector_ROC_MonteCarlo_sortxH0(t_mc[:, 0], t_mc[:, 2])
    Pfa_ACE, Pd_ACE = array_detector_ROC_MonteCarlo_sortxH0(t_mc[:, 1], t_mc[:, 3])
    return [Pfa_AMF, Pd_AMF], [Pfa_ACE, Pd_ACE]



import cv2
import numpy as np
def get_video_resolution(standard):
    resolutions = {
        "16K": (15360, 8640),
        "12K": (11520, 6480),
        "10K": (10240, 5760),
        "8K": (7680, 4320),
        "6K": (6144, 3160),
        "5K": (5120, 2880),
        "4K UHD": (3840, 2160),  # Commonly referred to as 4K
        "4K DCI": (4096, 2160),  # Digital Cinema Initiatives standard
        "1440p": (2560, 1440),  # Quad HD (QHD)
        "1080p": (1920, 1080),  # Full HD (FHD)
        "1080i": (1920, 1080),  # Interlaced Full HD
        "720p": (1280, 720),  # HD
        "576p": (720, 576),  # SD for PAL systems
        "480p": (854, 480),  # SD for NTSC systems
        "360p": (640, 360),  # Standard YouTube resolution
        "240p": (426, 240),  # Lower resolution for mobile or low bandwidth
        "144p": (256, 144),  # Minimum resolution for YouTube
        "NTSC": (720, 480),  # Standard definition in the USA
        "PAL": (720, 576),  # Standard definition in Europe and Asia
        "SVGA": (800, 600),  # Super Video Graphics Array
        "XGA": (1024, 768),  # Extended Graphics Array
        "WXGA": (1280, 800),  # Widescreen Extended Graphics Array
        "SXGA": (1280, 1024),  # Super Extended Graphics Array
        "HD+": (1600, 900),  # High Definition Plus
        "UXGA": (1600, 1200),  # Ultra Extended Graphics Array
        "FHD+": (2160, 1440),  # Full HD Plus
        "QHD+": (3200, 1800),  # Quad HD Plus
        "UHD+": (5120, 2880),  # Ultra HD Plus
    }
    
    return resolutions.get(standard, (None, None))
def create_grid_video(video_count, NW, NH, video_directory,fps=30, output_filename="grid_video_4k.avi",
                      output_width = 3840, output_height = 2160):

    # Initialize video readers
    videos = [cv2.VideoCapture(f"{video_directory}/radar_video_{i+1}.avi") for i in range(video_count)]

    # Read the first frame to get video dimensions
    ret, frame0 = videos[0].read()
    # fps = videos[0].get(cv2.CAP_PROP_FPS)
    if not ret:
        print("Failed to read video")
        return
    f0_flag = True
    videos[0].set(cv2.CAP_PROP_POS_FRAMES, 0)
    frame_height, frame_width = frame0.shape[:2]

    # Calculate new frame dimensions for grid layout
    tile_width = output_width // NW
    tile_height = output_height // NH

    # Initialize video writer
    fourcc = cv2.VideoWriter_fourcc(*'DIVX')
    out = cv2.VideoWriter(output_filename, fourcc, fps, (output_width, output_height))
    frameNumer = 0
    while True:
        frameNumer +=1
        # print(f'frameNumer = {frameNumer}')
        grid_frame = np.zeros((output_height, output_width, 3), dtype=np.uint8)
        all_videos_read = True

        for i in range(NH):
            for j in range(NW):
                idx = i * NW + j
                if idx >= video_count:
                    break
                if f0_flag:
                   f0_flag=False
                   ret, frame = True,frame0
                else:
                    ret, frame = videos[idx].read()
                if ret:
                    all_videos_read = False
                    resized_frame = cv2.resize(frame, (tile_width, tile_height))
                    grid_frame[i * tile_height:(i + 1) * tile_height, j * tile_width:(j + 1) * tile_width] = resized_frame

        if all_videos_read:
            break

        out.write(grid_frame)

    for video in videos:
        video.release()
    out.release()

def addframe_GridVideoWriters(images,videos_WH,videos):
  for i_image,image in enumerate(images):
    image = cv2.resize(image, (videos_WH[i_image][1], videos_WH[i_image][0]))
    videos[i_image].write(cv2.cvtColor(image, cv2.COLOR_RGB2BGR))

def firsttime_init_GridVideoWriters(images,video_directory,fps):
  videos = []
  videos_WH=[]

  for i_image,image in enumerate(images):
    frame_height, frame_width = image.shape[0], image.shape[1]  # Adjust as necessary
    videos_WH.append([frame_height, frame_width])
    #
    videos.append(cv2.VideoWriter(f"{video_directory}/radar_video_{i_image+1}.avi", cv2.VideoWriter_fourcc(*'DIVX'), fps, (frame_width, frame_height)))
  return videos,videos_WH  

def captureFig(fig):
  fig.canvas.draw()
  image = np.frombuffer(fig.canvas.tostring_rgb(), dtype='uint8')
  return image.reshape(fig.canvas.get_width_height()[::-1] + (3,))


from setuptools import setup, find_packages
from pathlib import Path

# Read the long description from the README.md file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding='utf-8')

# Define the required dependencies
install_requires = [
    'bpy>=4.0.0',                 # Blender Python API
    'matplotlib>=3.5.0',          # Plotting library
    'opencv-python>=4.5.0',       # OpenCV for computer vision
    'PyQt5>=5.15.0',              # PyQt5 for GUI applications
    'scipy>=1.9.0',               # Scientific computing
    'numba>=0.56.0',              # JIT compiler for Python
    'torch>=1.12.0',              # PyTorch for deep learning
    'scikit-learn>=1.0.0',        # Machine learning library
    'seaborn>=0.11.0',            # Statistical data visualization
    'tensorboard>=2.9.0',         # TensorBoard for visualizations
    'torchviz>=0.0.2',            # Visualization of PyTorch models
    'torchsummary>=1.5.0',        # Summary of PyTorch models
]

# Define the setup
setup(
    name='sensingsp',
    version='1.2.0',
    packages=find_packages(),
    install_requires=install_requires,
    url='https://sensingsp.github.io/',
    license='MIT',
    author='Moein Ahmadi',
    author_email='moein.ahmadi@uni.lu, gmoein@gmail.com',
    description=(
        'SensingSP™: Open-source library for simulating electromagnetic sensing systems '
        'and implementing signal processing and machine learning algorithms. Developed by '
        'the Signal Processing Applications in Radar and Communications (SPARC) group at '
        'the University of Luxembourg with support from IEE S.A.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    extras_require={
        'no_deps': [],  # Allows installing without any dependencies
    },
)
